/*
 * Created on 25-sep-2005
 *
 */
package parserString;

import java.util.Vector;

import terms.*;
/**
 * @author Rafael Caballero
 *
 * Converts a string in a Term. The string is supossed to contain a 
 * valid Term  
 **/
public class TermParser {

	/*
	 * Converts an string in a Term. If the string doesn't contain a term
	 * returns null
	 * @param s The string to be parsed
	 */
	public static Term parse(String s) {
	   Term result=null;
       StringIndex si = new StringIndex(s,0);
       si.skipBlanks();
       if (si.validIndex()) {
       	Term t1 = parse(si);
       	si.skipBlanks();
       	// if the string has been not completely parsed it must be an application
       	if (si.validIndex()) {
       		Vector v = new Vector();
       		while (si.validIndex()) {
       			Term t = parse(si);
       			v.addElement(t);
       			si.skipBlanks();
       		}
       		result = new Application(t1,v);
       	}
        else	result = t1;
       }	
       return result;
	}

	/**
	 * 
	 * @param s An string containg 0 or more terms separated by ,
	 * @return A vector with all the terms in the string
	 */
	
	
	/**
	 * Parses the next expression contained in a StringIndex 
	 * @param si The stringindex containing the string to be parsed.
	 *        The parsing process will start at the point of the index
	 *        and will end after having parsed an expression, leaving the 
	 *        index at the position of the last character of the expression
	 *        +1
	 * @return The term corresponding to the part of the string parsed,
	 *         or null is an error is found (it should not, we assume the 
	 *         string contains a valid term starting at the index position). 
	 *         
	 */
	public static Term parse(StringIndex si) {
		Term result = null;
		
		// first skip the blanks at the beginning
		si.skipBlanks();
		if (si.validIndex())
		  	// various possibilities
			switch (si.getChar()) {
		  	case '(':   // round bracket expression
		  		result = parseRoundBracketExp(si);
		  		break;
		  	case '[' : // list
		  		    result = parseList(si);
		  		    break;
		  	case '\'': // char
		  		    result = parseChar(si);
		  		    break;
		  	case '"': // string
		  		    result = parseString(si);
		  		    break;
		  	case '_': // bottom
		  		    si.incIndex();
		  		    if (si.validIndex() && Character.isLetter(si.getChar())) {
		  		    	si.decIndex();
		  		    	result = parseVariable(si);
		  		    } else 
		  		    	result = new Bottom();
		  		    break;	
		  		    		
		  	default:
			   if (Character.isUpperCase(si.getChar())) 
				  result = parseVariable(si);
				else if (Character.isDigit(si.getChar()))
					   result = parseNumber(si);
				else if (Character.isLowerCase(si.getChar()))
					  result = parseFunctor(si);
					 else if (operatorChar(si.getChar()))
					  result = parseOperator(si);	 
					 else result=null; // this case should not occurr		
		  	}          
			
		  return result;
		}

	
    
    /** parses a round bracktet expression. It can be either a tuple or a 
     * expression enclosed in round brackets. 
     * @param si An StringIndex with the index in a char with '('
     * @return A term representing the expression.
     *         The index in si will point to the first char after the ')' 
     */
    private static Term parseRoundBracketExp(StringIndex si) {
  		int pos = si.getIndex();
  		int posEnd= si.matchSymbol(')');
  		String s = si.subString(pos+1,posEnd);
  		si.setIndex(posEnd+1);
  		StringIndex s2 = new StringIndex( s,0 );
  		int posComma = s2.nextComma();
  		if (s2.validPosition(posComma)) {
  			// a tuple has the same form as a condition
  			Vector v = parseCondition(s);
  			return new Tuple(v);
  		}
  		else {	 
  			Term e = parse(s);
  			return new RoundBracketExpression(e);
  		}
    }

    public static Vector parseCondition(String s) {
    	Vector v = new Vector();
    	StringIndex si = new StringIndex(s,0);
    	while (si.validIndex()) {
    		// next condition
    		int start = si.getIndex();
    		// the end can be either the end of the expression or a ,
    		int end = si.nextComma();
    		
    		Term t =  si.validIndex() ? parse(si.subString(start,end))
    				               : parse(si.subString(start));
    		v.addElement(t);
    		si.incIndex();
    	}
    	return v;
    }
    /**
     * 
     * @param si Input string. The index must correspond to a '[' character
     * @return A term representing the list. The 
     *         The index in si will be point to the first char after the ']' 
     */
   private static Term parseList(StringIndex si) {
     Term result = null;
     
     int start = si.getIndex();
     int end = si.matchSymbol(']');
     
     StringIndex sList = new StringIndex(si.subString(start+1,end),0);
     
     // change the index string to the first position after the ]
     si.setIndex(end+1);
     
     // a list can be either a sequence of elements or a list with 
     // head and tail
     int indexPipe = sList.next('|');
     
     if (sList.validPosition(indexPipe)) {
     	// head and tail
     	String sHead = sList.subString(0,indexPipe);
     	String sTail = sList.subString(indexPipe+1);
     	Vector head = parseCondition(sHead);
     	Term tail = parse(sTail);
     	result = new HeadTailList(head,tail);
     	
     } else {
     	sList.skipBlanks();
     	Vector v = parseCondition(sList.getString());
     	result = new SeqList(v);
     }
     
     return result;
   }
   
   private static Term parseNumber(StringIndex si) {
   	Term result = null;
   	boolean isReal = false;
   	
   	// number can be either an integer or a real
   	int start = si.getIndex();
   	while (si.validIndex() && Character.isDigit(si.getChar()) )
   		   si.incIndex();

	String s1 = si.validIndex() ? si.subString(start,si.getIndex())
                              : si.subString(start);
    
   	if (si.validIndex() && si.getChar()=='.') {
   		si.incIndex();
   		if (si.validIndex() && Character.isDigit(si.getChar()))
   			isReal = true;
   		else
   			si.decIndex();
   	}
   	
   	if (isReal) {
   		start = si.getIndex();
   		while (si.validIndex() && Character.isDigit(si.getChar()) )
    		   si.incIndex();
		String s2 = si.validIndex() ? si.subString(start,si.getIndex())
                                   : si.subString(start);

 		result = new ToyReal(s1,s2);
   	}
   	else 
   		result = new ToyInt(s1);
   		 
   	return result;
   }

   private static Term parseChar(StringIndex si) {
   	Term result = null;
   	si.incIndex();
   	char c = si.getChar();
   	si.decIndex();
   	si.skipChar();
   	si.incIndex();
   	result = new ToyChar(c);
       	
   	return result;
   }

   private static Term parseString(StringIndex si) {
   	Term result = null;
   	int start = si.getIndex();
   	si.skipString();
   	String s = si.validIndex() ?  si.subString(start+1,si.getIndex())
   			                   : si.subString(start+1);
   	result = new ToyString("\""+s+"\""); // rafa 30-12-05: introduced the "\""
    si.incIndex();   	
   	return result;
   }

   private static Term parseVariable(StringIndex si) {
   	Term result = null;
   	int start = si.getIndex();
   	while (si.validIndex() &&
   			Character.isJavaIdentifierPart( si.getChar() ) )
   		  si.incIndex();
   	String id = "";
   	
   	id = si.validIndex() ? si.subString(start,si.getIndex())
   				             : si.subString(start);
   	
   	result = new Variable(id);
   	return result;
   }

   private static Term parseFunctor(StringIndex si) {
   	Term result = null;
   	int start = si.getIndex();
   	while (si.validIndex() &&
   			Character.isJavaIdentifierPart( si.getChar() ) )
   		  si.incIndex();
   	String id = "";
   	
   	id = si.validIndex() ? si.subString(start,si.getIndex())
   				             : si.subString(start);
   	
   	result = new Functor(id);
   	return result;
   }

   /**
    * Parses a Toy operator. 
    * 
    * @param si The string to be parsed. Precond. operatorChar(si.getChar())==true 
    * @return A term representing the operator. The index in si is advanced to
    *  one position beyond the end of the operator
    */
   private static Term parseOperator(StringIndex si) {
   	Term result = null;
   	int start = si.getIndex();
   	while (si.validIndex() && operatorChar( si.getChar() ) )
   		  si.incIndex();
   	String id = "";
   	
   	id = si.validIndex() ? si.subString(start,si.getIndex())
   				             : si.subString(start);
   	
   	result = new Operator(id);
   	return result;
   }

   /**
    * Checks wether a given character can be part of a Toy operator or not. 
    * @param c Charactger to be checked.
    * @return true if the character can be part of a Toy operator. false otherwise.
    */
   private static boolean operatorChar(char c) {
	return c=='!'   || c=='#' || c=='&'  || c=='|' || c=='/' ||
			c=='\\' || c=='*' || c=='+' || c=='-' || c=='.'  ||
   			c==':'  || c=='<' || c=='>' || c=='=' || c=='@'  ||
   			c=='~'  || c=='%' || c=='' || c=='?' || c=='$';
   }

}
