/* Copyright(C) 1999, Swedish Institute of Computer Science */
/***
  Support for all_distinct/[1,2], assignment/[2,3], circuit/[1,2].
  Algorithms:
  Regin '94, Costa '94, Caseau & Laburthe '97, Mehlhorn & Thiel '00,
  and more for circuit.
  ***/

#include "fd.h"
#include "dvars.h"

#define CSTATE_ONLY 0

/* Optimization: avoid computing SCCs if the previously
   computed SCCs are still valid, and the same number of vars are
   still reachable from exposed values.  Does not pay off.

   Some statistics on SCC computation:
   Case 1: all vars reachable.
   Case 2: all vars fixed or reachable.
   Case 3: fast_find_sccs.
   Case 4: single_scc.
   Case 5: general case.

   Case          :    1    2    3    4    5
   ========================================
   Social golfers: 0.00 0.01 0.99 0.00 0.00
   Golomb 10     : 0.55 0.31 0.02 0.00 0.11
    8 Queens     : 0.17 0.68 0.03 0.06 0.06
   10 Queens     : 0.17 0.66 0.02 0.06 0.09
   12 Queens     : 0.17 0.66 0.02 0.05 0.10
*/
#define WITNESS 0

/* Optimization: avoid recomputing the set of reachable
   (from exposed values) variables, by checking if the
   previous set is still valid.  For Golomb, this succeeded
   in 20% of the cases and failed in 80%.
   Does not seem to pay off.
*/
#define INCTREE 0

#if MULTI_SP_AWARE
#if 0                           /* not redefined. It uses FD_SETUP_SPENV to access the SPEnv */
#define all_distinct_destructor(A1) all_distinct_destructor(HIDDEN_ARG, A1)
#endif /* 0 */

#define qsort_asc_longswap(A1,A2,A3,A4) qsort_asc_longswap(HIDDEN_ARG, A1,A2,A3,A4)
#define qsort_asc_longmed3(A1,A2,A3) qsort_asc_longmed3(HIDDEN_ARG, A1,A2,A3)
#define qsort_asc_long(A1,A2) qsort_asc_long(HIDDEN_ARG, A1,A2)
#define cmp_asc_val(A1,A2) cmp_asc_val(HIDDEN_ARG, A1,A2)
#define qsort_asc_valswap(A1,A2,A3,A4) qsort_asc_valswap(HIDDEN_ARG, A1,A2,A3,A4)
#define qsort_asc_valmed3(A1,A2,A3) qsort_asc_valmed3(HIDDEN_ARG, A1,A2,A3)
#define qsort_asc_val(A1,A2) qsort_asc_val(HIDDEN_ARG, A1,A2)
#if 0 /* MAY COME IN HANDY */
#define cmp_asc_cost(A1,A2) cmp_asc_cost(HIDDEN_ARG, A1,A2)
#define qsort_asc_costswap(A1,A2,A3,A4) qsort_asc_costswap(HIDDEN_ARG, A1,A2,A3,A4)
#define qsort_asc_costmed3(A1,A2,A3) qsort_asc_costmed3(HIDDEN_ARG, A1,A2,A3)
#define qsort_asc_cost(A1,A2) qsort_asc_cost(HIDDEN_ARG, A1,A2)
#endif /* 0 */
#define clear_mate(A1) clear_mate(HIDDEN_ARG, A1)
#define remove_neighbor(A1,A2) remove_neighbor(HIDDEN_ARG, A1,A2)
#define neighbor_of(A1,A2) neighbor_of(HIDDEN_ARG, A1,A2)
#define propagate_value(A1,A2) propagate_value(HIDDEN_ARG, A1,A2)
#define augment_path(A1,A2) augment_path(HIDDEN_ARG, A1,A2)
#define augment(A1) augment(HIDDEN_ARG, A1)
#define fast_find_sccs(A1) fast_find_sccs(HIDDEN_ARG, A1)
#define visit(A1,A2,A3) visit(HIDDEN_ARG, A1,A2,A3)
#define find_sccs(A1) find_sccs(HIDDEN_ARG, A1)
#define single_scc(A1) single_scc(HIDDEN_ARG, A1)
#define compress_edges(A1,A2) compress_edges(HIDDEN_ARG, A1,A2)
#define all_dist_filter(A1) all_dist_filter(HIDDEN_ARG, A1)
#define ass_filter(A1) ass_filter(HIDDEN_ARG, A1)
#define remove_edge(A1,A2) remove_edge(HIDDEN_ARG, A1,A2)
#define single_circuit(A1,A2,A3) single_circuit(HIDDEN_ARG, A1,A2,A3)
#define all_dist_alloc(A1,A2,A3,A4,A5,A6) all_dist_alloc(HIDDEN_ARG, A1,A2,A3,A4,A5,A6)
#define all_dist_alloc_temp(A1) all_dist_alloc_temp(HIDDEN_ARG, A1)
#define all_dist_free_temp(A1) all_dist_free_temp(HIDDEN_ARG, A1)
#define cost_filter(A1,A2,A3) cost_filter(HIDDEN_ARG, A1,A2,A3)
#define cost_based_pruning(A1,A2,A3) cost_based_pruning(HIDDEN_ARG, A1,A2,A3)
#define cons_batch(A1,A2) cons_batch(HIDDEN_ARG, A1,A2)
#define pairing_filter(A1) pairing_filter(HIDDEN_ARG, A1)


#if DynamicMallocBase /* [PM] 3.10.1 MAGIC needed for DynamicMallocBase */
#define all_dist_init(A1,A2) all_dist_init(HIDDEN_ARG, A1,A2)
#define ass_init(A1,A2) ass_init(HIDDEN_ARG, A1,A2)
#define all_pairs_shortest_paths(A1,A2) all_pairs_shortest_paths(HIDDEN_ARG, A1,A2)
#define hungarian_init(A1,A2,A3) hungarian_init(HIDDEN_ARG, A1,A2,A3)
#define hungarian_body(A1,A2,A3,A4,A5) hungarian_body(HIDDEN_ARG, A1,A2,A3,A4,A5)
#define cost_based_filter(A1,A2) cost_based_filter(HIDDEN_ARG, A1,A2)
#endif /* DynamicMallocBase */

#endif /* MULTI_SP_AWARE */

typedef long VERTEX;

#define EOL (-1L)

struct all_dist_data {
  void (SPCDECL *destructor)(void*);
#if MULTI_SP_AWARE
  SPEnv *spenv;
#endif /* MULTI_SP_AWARE */
  SP_globref refbase;
  long stamp;			/* increases up to backtracking */
  BOOL dense;			/* TRUE if vars 1..n, vals 1..n */
  BOOL pruned;
  BOOL incremental;
  int nrefs;
  int nvars;
  int nvartargets;
#if CSTATE_ONLY
  int nvars_back;
  int nvals_back;
#endif
  int nvals;
  int nvaltargets;
  int slack;			/* #vals - #vars after filtering */
  int scc_component;		/* SCC number */
  int scc_visited;		/* Vertex number during SCC search */
  int scc_index;		/* Stack ptr during SCC search */
  int reachable;		/* #vars reachable from exposed vals */
  Dvar dvar;
  VERTEX *vartarget;		/* var vertices: nvars*2 */
  VERTEX *valtarget;		/* val vertices: nvals*2 */
  VERTEX *dist_stack;		/* : 3*nvars+nvals, volatile */
#if WITNESS
  int     reachwit;		/* saved reachable from last time */
  long   *sccwitix;		/* next chunk in sccwit, nvars */
  VERTEX *sccwit;		/* recorded SCCS, nvars */
#endif
#if INCTREE
  int nreach;			/* #vars reachable from exposed vals */
  VERTEX *reachset;		/* recorded reachable, nvars */
#endif
  struct {
    TAGGED *val;		/* The value of the vertex.  1-based: nvars+nvals  */
    VERTEX *mate;		/* Mate of this vertex */
    long *component;		/* Component number */
    long *live;			/* Degree i.e. next index in neighbor array */
    long *size;			/* Size of domain (needed for pruning), volatile */
    long *visited;		/* VAL nodes: Nth visited in SCC search */
				/* VAR nodes: visited in augment_path */
  } vertex;			/* : nvars+nvals */
  /* additions for cost_based pruning */
  long *cost;			/* matrix if cost-based, else NULL */
  struct {
    /* the following fields are valid while descending a branch */
    long costbase;		/* cost of all ground items */
    int ntargets;		/* rc indexes range in [0,ntargets-1] */
    int *index_var;		/* rc index to vertex no. */
    int *var_index;		/* vertex no. to rc index */
    long *lbdual[2];		/* min. cost dual soln. */
    long *ubdual[2];		/* max. cost dual soln. */
    int  *lbprimal[2];		/* min. cost primal soln. */
    int  *ubprimal[2];		/* max. cost primal soln. */
    Dvar costvar;
    long *rcost[2];		/* 0=min, 1=max */
    long optcost[2];		/* 0=min, 1=max */
    /* the following fields are volatile */
    int *label[2];		/* for finding aug. path */
    long *cost;
    long *dist;
    long *distrow;		/* points at k'th row of dist */
    int *queue;
    int *index;
    int *heap;
    int *vheap;
    int heapsize;
    long min_regret[2];		/* 0=min, 1=max */
    long max_regret[2];		/* 0=min, 1=max */
    long threshold[2];		/* 0=min, 1=max */
  } rc;
};

#define DVAR(V) (pdata->dvar+(V))
#define VAL(V) (pdata->vertex.val[V])
#define SIZE(V) (pdata->vertex.size[V])
#define NEIGHS(V) (pdata->vertex.neighs[V])
#define MATE(V) (pdata->vertex.mate[V])
#define LIVE(V) (pdata->vertex.live[V])
#define COMPONENT(V) (pdata->vertex.component[V])
#define VISITED(V) (pdata->vertex.visited[V])
#define RefAttr(V) (pdata->refbase + 2*(V))
#define RefVar(V) (pdata->refbase + 2*(V) + 1)
#define TVAR(V) (pdata->vartarget[V])
#define TVAL(V) (pdata->valtarget[V])

/* Invariant: VISITED(V)>EOL iff V is a target (not yet ground and checked) */

static BOOL cost_filter MAGIC (HIDDEN_PROTO
			       struct all_dist_data *pdata,
			       long costbase,
			       int flags);

static void SPCDECL all_distinct_destructor(void *pdata_v)
{
  struct all_dist_data *pdata = (struct all_dist_data *)pdata_v;
  FD_SETUP_SPENV(pdata->spenv)

  SP_free_globrefs(pdata->refbase,pdata->nrefs);
  SP_free(pdata);
}

static int cmp_asc_val MAGIC (HIDDEN_PROTO
			      VERTEX *t1, VERTEX *t2)
{
  struct all_dist_data *pdata = fd.gdata;
  TAGGED val1 = VAL(*t1);
  TAGGED val2 = VAL(*t2);

  return TCMP(val1,val2);
}

#define QType VERTEX
#define QCmp  cmp_asc_val
#define QSort qsort_asc_val
#include "qsort.ic"

static INLINE void clear_mate MAGIC (HIDDEN_PROTO
                                     VERTEX vertex)
{
  struct all_dist_data *pdata = fd.gdata;
  VERTEX mate = MATE(vertex);

  if (mate>EOL)
    MATE(mate) = MATE(vertex) = EOL;
}

#define set_mates(n1,n2) MATE(n1) = (n2), MATE(n2) = (n1)

#define exposed(vertex) (MATE(vertex) == EOL)

static int remove_neighbor MAGIC (HIDDEN_PROTO
				  VERTEX u, VERTEX v)
{
  struct all_dist_data *pdata = fd.gdata;
  TAGGED tv = VAL(v);
  Dvar udv = DVAR(u);

  if (dvar_prune_value_t(udv,tv)) {
    LIVE(u)--;
    pdata->pruned = TRUE;
  }

  return LIVE(u);
}

static VERTEX neighbor_of MAGIC (HIDDEN_PROTO
                                 TAGGED tval, VERTEX u)
{
  struct all_dist_data *pdata = fd.gdata;

  if (u>=pdata->nvars)
    return GetSmall(tval)-1;
  else if (pdata->dense)
    return GetSmall(tval)+pdata->nvars-1;
  else {
    /* dichotomic search for j such that VAL(TVAL(j))==tval */
    int j, mid, sup;
    
    j = 0;
    sup = pdata->nvaltargets;
    while (j<sup) {
      mid = (j+sup)>>1;
      if (Tlt(VAL(TVAL(mid)),tval)) /* ORDER */
	j = mid+1;
      else
	sup = mid;
    }
    return TVAL(j);
  }
}

static BOOL propagate_value MAGIC (HIDDEN_PROTO
				   int nvars, int nvals)
{/* Performs all possible pruning by propagating singletons.
    Also, build matching for such singletons. */
  int i, top=0;
  struct all_dist_data *pdata = fd.gdata;
  VERTEX *dist_stack = pdata->dist_stack;
  DVITER it;

  for (i=0; i<nvars; i++) {
    VERTEX v = TVAR(i);
     
    if (LIVE(v) == 1)
      dist_stack[top++] = v;
  }
  for (i=0; i<nvals; i++) {
    VERTEX v = TVAL(i);
       
    if (LIVE(v) == 1)
      dist_stack[top++] = v;
  }

  while (top>0) {
    VERTEX vertex = dist_stack[--top];
    VERTEX rvertex = neighbor_of(dvar_min_t(DVAR(vertex)), vertex);
    TAGGED tv = VAL(vertex);
    Dvar rdv = DVAR(rvertex);

    if (MATE(vertex) != rvertex) {
      clear_mate(vertex);
      clear_mate(rvertex);
      set_mates(vertex,rvertex); /* one matching done */
    }
    if (LIVE(rvertex) > 1) {
      dviter_init(&it, rdv);
      while (!dviter_empty(&it)) {
	VERTEX lvertex = neighbor_of(dviter_next_value_t(&it), rvertex);
	
	if (lvertex!=vertex)
	  switch (remove_neighbor(lvertex,rvertex)) {
	  case 0: return FALSE;
	  case 1: dist_stack[top++] = lvertex;
	  }
      }
      dvar_fix_value_t(rdv, tv);
      LIVE(rvertex) = 1;
      pdata->pruned = TRUE;
    }
  }

  return TRUE;
} /* propagate_value */

static BOOL augment_path MAGIC (HIDDEN_PROTO 
				VERTEX vertex,
				int now)
   /* BFS algorithm for finding an augmenting path. 
      If successful, updates the graph to reflect the new matching.
      Return TRUE or FALSE, depending on whether an augmenting path is
      found.
      */
{
  int p=0, q=0;
  VERTEX newvertex, neighbor;
  struct all_dist_data *pdata = fd.gdata;
  VERTEX *vstack = pdata->dist_stack;
  DVITER it;

  vstack[q++] = vertex;
  while (p<q) {
    vertex = vstack[p++];
    dviter_init(&it, DVAR(vertex));
    while (!dviter_empty(&it)) {
      neighbor = neighbor_of(dviter_next_value_t(&it), vertex);
      newvertex = MATE(neighbor);
      if (newvertex==EOL) {
	for (;;) {
	  newvertex = MATE(vertex);
	  set_mates(vertex,neighbor);
	  if (newvertex==EOL)
	    return TRUE;
	  neighbor = newvertex;
	  vertex = vstack[newvertex]; /* FATHER(newvertex) */
	}
      } else if (VISITED(newvertex)!=now) {
	VISITED(newvertex) = now;
	vstack[neighbor] = vertex; /* FATHER(neighbor) */
	vstack[q++] = newvertex;
      }
    }
  }
  return FALSE;
} /* augment_path */



static BOOL augment MAGIC (HIDDEN_PROTO
			   int nvars)
   /* This function returns TRUE and a new graph with a maximum matching
      if there is such a matching. Returns FALSE otherwise. The maximum
      matching is found by repeatedly finding augmenting paths in the
      graph until all variable vertices are matched.  */
{
  int i, first_exp;
  VERTEX vertex, rvertex;
  TAGGED seed=0;		/* avoid false alarm */
  struct all_dist_data *pdata = fd.gdata;
				/* greedy phase */
  first_exp = nvars;
  for (i=0; i<nvars; i++) {
    vertex = TVAR(i);
    if (exposed(vertex)) {
      if (first_exp>i) {
	first_exp = i;
	seed = dvar_min_t(DVAR(vertex));
	rvertex = neighbor_of(seed,vertex);
	if (exposed(rvertex))
	  set_mates(vertex,rvertex);
      } else {
	seed = dvar_successor_t(DVAR(vertex),seed);
	if (TagIsSmall(seed)) {
	  rvertex = neighbor_of(seed,vertex);
	  if (exposed(rvertex))
	    set_mates(vertex,rvertex);
	}
      }
    }
  }
				/* augmenting phase */
  for (i=first_exp; i<nvars; i++) {
    vertex = TVAR(i);
    if (exposed(vertex)) {
      if (!augment_path(vertex,i+1))
	return FALSE;
    }
  }
  return TRUE;
} /* augment */


/* Returns number of SCCs found. */
static int
fast_find_sccs MAGIC (HIDDEN_PROTO
		      int sccid)
{
  int sccid0 = sccid;
  int j, top1, top2;
#if WITNESS
  int ix1, ix2;
#endif
  struct all_dist_data *pdata = fd.gdata;
  int numberx = pdata->nvars;
  int numbery = pdata->nvaltargets;
  VERTEX aftery = numberx + pdata->nvals;
  struct {
    VERTEX *node;
    VERTEX *root;
    VERTEX *rightmost;
    VERTEX *maxx;
  } stk;

  /* init */
  stk.node      = pdata->dist_stack;
  stk.root      = stk.node+numberx;
  stk.rightmost = stk.root+numberx;
  stk.maxx      = stk.rightmost+numberx;
#if WITNESS
  ix1=0; ix2=0;
#endif
  j=0; top1=0; top2=0;
  while (j-top1<numbery) {
    VERTEX y = j<numbery ? TVAL(j) : aftery;
    VERTEX x = j<numbery ? MATE(y) : numberx;
    if (j<numbery && (x==EOL || VISITED(y))) /* node to be ignored */
      j++;
    else if (top2==0) {	/* start 1st component */
      stk.node[0] = y;
      stk.root[0] = y;
      stk.rightmost[0] = y;
      stk.maxx[0] = neighbor_of(dvar_max_t(DVAR(x)), x); /* SHAVE? */
      top1 = top2 = 1;
      j++;
    } else if (stk.maxx[top2-1]>=y) { /* start/extend component */
      VERTEX min1 = neighbor_of(dvar_min_t(DVAR(x)), x);
      VERTEX max1 = neighbor_of(dvar_max_t(DVAR(x)), x);
      
      stk.node[top1++] = y;
      stk.root[top2] = y;
      while (top2>0 && min1<=stk.rightmost[top2-1]) { /* merge components */
	top2--;
	if (max1<stk.maxx[top2])
	  max1 = stk.maxx[top2];
      }
      stk.rightmost[top2] = y;
      stk.maxx[top2++] = max1; /* SHAVE? */
      j++;
    } else {			/* component done */
      VERTEX root1 = stk.root[--top2];
      VERTEX x1, y1;

      sccid++;
      do {
	y1 = stk.node[--top1];
	x1 = MATE(y1);
	COMPONENT(y1) = sccid;
	COMPONENT(x1) = sccid;
#if WITNESS
	pdata->sccwit[ix2++] = x1;
#endif
      } while (y1>root1);
#if WITNESS
      pdata->sccwitix[ix1++] = ix2;
#endif
    }
  }
#if WITNESS
  pdata->sccwitix[ix1++] = EOL;
#endif
  return sccid - sccid0;
}

static int visit MAGIC (HIDDEN_PROTO
			VERTEX vertex, int *ix1, int *ix2)
   /* A recursive function that finds strongly connected components
      [Tarjan'72]. Code taken roughly from [Sedgewick, Algorithms in C,
      page 482].
      One big difference is that we only step on value vertices, i.e. we
      jump directly from variable vertices using the mate ptr.
      */
{
  VERTEX newvertex=EOL, mate;
   int m, min, scc;
   struct all_dist_data *pdata = fd.gdata;
   VERTEX *dist_stack = pdata->dist_stack;
   DVITER it;
   
   VISITED(vertex) = ++pdata->scc_visited;		/* Mark vertex as the id'th visited */
   min = pdata->scc_visited;
   dist_stack[pdata->scc_index++] = vertex;
   mate = MATE(vertex);
   dviter_init(&it, DVAR(mate));
   while (!dviter_empty(&it)) {
     newvertex = neighbor_of(dviter_next_value_t(&it),mate);
     if (newvertex != vertex) {
       m = (VISITED(newvertex) == 0) ? 
	 visit(newvertex,ix1,ix2) : VISITED(newvertex);
       if (m < min)
	 min = m;
     }
   }
   if (min == VISITED(vertex)) {
     scc = ++pdata->scc_component;
     do {
       VERTEX mate;
       /* Each descendant on the dist_stack is part of this vertex's SCC. */
       newvertex = dist_stack[--pdata->scc_index];
       mate = MATE(newvertex);
       COMPONENT(newvertex) = scc;
       COMPONENT(mate) = scc;
#if WITNESS
       pdata->sccwit[(*ix2)++] = mate;
#endif
       VISITED(newvertex)=0xffffff; /* High value, so that
				       this vertex will be ignored
				       in the future search. */
     } while (newvertex != vertex);
#if WITNESS
     pdata->sccwitix[(*ix1)++] = *ix2;
#endif
   }
   return min;
} /* visit */


static BOOL single_scc MAGIC (HIDDEN_PROTO
			      TAGGED unvisited)
{
  TAGGED y, preds, fdset;
  VERTEX yvertex;
  int cmp = FDI_DISJOINT;
  struct all_dist_data *pdata = fd.gdata;
#if WITNESS
  int ix2 = 0;
#endif

  y = fd_min(unvisited);
  preds = fd_interval(y,y);
  yvertex = neighbor_of(y,0);
  unvisited = fd_delete(unvisited,y);
#if WITNESS
  pdata->sccwit[ix2++] = MATE(yvertex);
#endif
  fdset = dvar_set(DVAR(MATE(yvertex)));
  while (unvisited!=EmptySet) {
    y = fd_and_min(unvisited,fdset);
    if (y==ERRORTAG)
      return FALSE;
    unvisited = fd_delete(unvisited,y);
    yvertex = neighbor_of(y,0);
    fdset = dvar_set(DVAR(MATE(yvertex)));
    cmp = fd_compare(fdset,preds);
    if (cmp!=FDI_DISJOINT)
      preds = fd_or_interval(preds,y,y);
#if WITNESS
    pdata->sccwit[ix2++] = MATE(yvertex);
#endif
  }
#if WITNESS
  pdata->sccwitix[0] = ix2;
  pdata->sccwitix[1] = EOL;
#endif
  return (cmp != FDI_DISJOINT);
}

/* Returns #non-singleton components. */
static int find_sccs MAGIC (HIDDEN_PROTO  /* (sccs,nvars) */
			    int sccs)
   /* Marks all edges belonging to any strongly connected component in
      the graph. The strongly connected components are calculated using
      Tarjan's depth-first search based algorithm from 1972. */
{
   int i, ix1, ix2;
   FDCONS cons;
   VERTEX vertex, mate;
   Dvar dv;
   struct all_dist_data *pdata = fd.gdata;
   int nvars = pdata->nvartargets;
   int nvals = pdata->nvaltargets;
     
#if WITNESS
   if (pdata->incremental && pdata->reachwit==pdata->reachable) {
     BOOL valid = TRUE;
     ix1 = ix2 = 0;
     while (valid && pdata->sccwitix[ix1]!=EOL) {
       int lim = pdata->sccwitix[ix1++];
       VERTEX head = EOL;
       VERTEX cur;

       while (valid && ix2<lim) {
	 vertex = pdata->sccwit[ix2++];
	 dv = DVAR(vertex);
	 if (VISITED(vertex)>EOL && !dvar_is_integer(dv)) {
	   if (head==EOL)
	     head = vertex;
	   else if (valid)
	     valid = dvar_contains_value_t(dv, VAL(MATE(cur)));
	   cur = vertex;
	 }
       }
       if (head!=EOL && valid) {
	 valid = dvar_contains_value_t(DVAR(head), VAL(MATE(cur)));
       }
     }
     if (valid)
       return 0;
   }
   pdata->reachwit = pdata->reachable;
#endif
   fdcons_init(&cons);
   for (i=0; i<nvals; i++) {
     vertex = TVAL(i);
     mate = MATE(vertex);
     if (mate==EOL)
       VISITED(vertex) = 0xffffff; /* ignore in SCC search */
     else if (COMPONENT(mate)!=0) /* reachable from exposed node */
       VISITED(vertex) = 0xffffff; /* ignore in SCC search */
     else if (LIVE(mate)==1) {
       VISITED(vertex) = 0xffffff; /* ignore in SCC search */
       COMPONENT(mate) = ++sccs; /* singleton SCC */
       COMPONENT(vertex) = sccs;
     } else
       fdcons_add(&cons,VAL(vertex)); /* ORDER */
   }
   if (fdcons_size(&cons)==0) {	/* all vars fixed or reachable */
     return 0;
   }
   for (i=0; i<nvars; i++) {
     vertex = TVAR(i);
     mate = MATE(vertex);
     dv = DVAR(vertex);
     if (!VISITED(mate)) {
       TAGGED holes =
	 fd_interval_subtract(dvar_min_t(dv), dvar_max_t(dv), dvar_set(dv));

       if (holes != EmptySet &&
	   fd_compare(holes,fdcons_set(&cons))!=FDI_DISJOINT)
	 goto not_convex;
     }
   }
   return fast_find_sccs(sccs/* latest used */);
 not_convex:
   if (single_scc(fdcons_set(&cons))) {
     return 1;
   }
   /* Golomb 10: here with 6% prob. */
   /* 12 Queens: here with 66% prob. */
   pdata->scc_component = sccs;
   pdata->scc_visited = 0;
   pdata->scc_index = 0;
   ix1 = ix2 = 0;
   for (i=0; i<nvars; i++) {
     vertex=MATE(TVAR(i));
     if (VISITED(vertex) == 0)
       visit(vertex,&ix1,&ix2);
   }
#if WITNESS
   pdata->sccwitix[ix1++] = EOL;
#endif
   return pdata->scc_component - sccs;
} /* find_sccs */


static void compress_edges MAGIC (HIDDEN_PROTO
				  VERTEX *vertices,
				  int nvars)
{
  struct all_dist_data *pdata = fd.gdata;
  int i, size;
  DVITER it;
  FDCONS cons;
  
  for (i=0; i<nvars; i++) {
    VERTEX rvertex;
    VERTEX vertex = vertices[i];
    Dvar dv = DVAR(vertex);
    int component = COMPONENT(vertex);

    dviter_init(&it, dv);
    fdcons_init(&cons);
    while (!dviter_empty(&it)) {
      rvertex = neighbor_of(dviter_next_value_t(&it),vertex);
      if (COMPONENT(rvertex) != component)
	fdcons_add(&cons, VAL(rvertex));
    }
    size = fdcons_size(&cons);
    if (size>0) {
      dvar_prune_set(dv, fdcons_set(&cons));
      pdata->pruned = TRUE;
      LIVE(vertex) -= size;
    }
  }
}

static BOOL all_dist_filter MAGIC (HIDDEN_PROTO
				   struct all_dist_data *pdata)
{
  int i, j, rcall;
  VERTEX vertex, rvertex;
  int nvartargets = pdata->nvartargets;
  int nvaltargets = pdata->nvaltargets;
  int slack = nvaltargets-nvartargets;
  int nunreach = nvartargets;
  
  if (slack<0)
    return FALSE;

  for (i=0; i<nvartargets; i++) {
    vertex = TVAR(i);
    VISITED(vertex) = 0;
    COMPONENT(vertex) = 0;
  }
  for (i=0; i<nvaltargets; i++) {
    vertex = TVAL(i);
    VISITED(vertex) = 0;
    /* COMPONENT is used in the incremental case */
  }
  do {
    int nint = 0;
    int rc;
    Dvar dv;

    rcall = 0;    
    for (i=0; i<nvartargets; i++) {
      vertex = TVAR(i);
      if (LIVE(vertex)==1) {
	rvertex = neighbor_of(dvar_min_t(DVAR(vertex)),vertex);
	VISITED(vertex) = 1;
	if (VISITED(rvertex)++ != 0)
	  return FALSE;
	LIVE(rvertex) = 1;	/* used in SCC search */
	if (MATE(vertex) != rvertex) {
	  clear_mate(vertex);
	  clear_mate(rvertex);
	  set_mates(vertex,rvertex); /* one matching done */
	}
	nint++;
      }
    }
    if (nint>0) {
      FDCONS cons;
      fdcons_init(&cons);
      for (i=0; i<nvaltargets; i++) {
	vertex = TVAL(i);
	if (VISITED(vertex)!=0) {
	  VISITED(vertex) = 0;
	  fdcons_add(&cons, VAL(vertex));
	}
      }
      for (i=0; i<nvartargets; i++) {
	vertex = TVAR(i);
	dv = DVAR(vertex);
	if (VISITED(vertex)!=0) {
	  VISITED(vertex) = 0;
	} else {
	  rc = dvar_prune_set(dv, fdcons_set(&cons));
	  rcall |= rc;
	  if (rc<0)
	    return FALSE;
	  LIVE(vertex) = dvar_value_count(dv);
	  pdata->pruned = TRUE;
	}
      }
    }
  } while (rcall & DV_PRUNED_VAL);
  pdata->slack = slack;

  /* Find maximal matching. */
  if (!augment(nvartargets))
    return FALSE;
  
  /* Look for all even length paths starting at an exposed value vertex. */
  if (slack==0) {
    for (i=0; i<nvaltargets; i++) {
      vertex = TVAL(i);
      COMPONENT(vertex) = 0;
    }
  } else {
    FDCONS reach0;
    TAGGED reachable;
    VERTEX mate;
    VERTEX *cands = pdata->dist_stack;
    VERTEX *src = pdata->valtarget;
    int nr;
    long comp;

    fdcons_init(&reach0);
    for (i=j=0; i<nvaltargets; i++) {
      vertex = src[i];
      mate = MATE(vertex);
      comp = COMPONENT(vertex);
      COMPONENT(vertex) = 0;
      if (exposed(vertex)) {
	COMPONENT(vertex) = 1;
	fdcons_add(&reach0, VAL(vertex)); /* ORDER */
      } else if (LIVE(mate)>1) {
	if (!pdata->incremental || comp==1)
	  cands[j++] = vertex;
      }
    }
    nvaltargets = j;

#if INCTREE
    if (pdata->incremental) {
      reachable = fdcons_set(&reach0);
      nr = 0;
      for (i=0; i<pdata->nreach; i++) {
	vertex = pdata->reachset[i];
	if (LIVE(vertex)>1) {
	  if (dvar_compare_set(DVAR(vertex),reachable)==FDI_DISJOINT) {
	    goto invalid;
	  }
	  pdata->reachset[nr++] = vertex;
	  reachable = fd_insert_into(VAL(MATE(vertex)), reachable);
	}
      }
      for (i=0; i<nr; i++) {
	vertex = pdata->reachset[i];
	COMPONENT(vertex) = 1;
	COMPONENT(MATE(vertex)) = 1;
      }
      goto finish;
    }
  invalid:
#endif
    reachable = EmptySet;
    nr = 0;
    while (fdcons_set(&reach0) != EmptySet) {
      reachable = fd_or(reachable,fdcons_set(&reach0));
      fdcons_init(&reach0);
      for (i=j=0; i<nvaltargets; i++) {
	vertex = cands[i];
	mate = MATE(vertex);
	if (dvar_compare_set(DVAR(mate),reachable)!=FDI_DISJOINT) {
	  COMPONENT(vertex) = 1;
	  COMPONENT(mate) = 1;
	  fdcons_add(&reach0, VAL(vertex)); /* ORDER */
#if INCTREE
	  pdata->reachset[nr] = mate;
#endif
	  nr++;
	} else
	  cands[j++] = vertex;
      }
      nvaltargets = j;
    }
#if INCTREE
  finish:
    pdata->nreach = nr;
#endif
    nunreach -= nr;
    if (nunreach==nvartargets)	/* all exposed vertices are dead */
      pdata->slack = 0;
  }
  pdata->reachable = nvartargets-nunreach;
  
  /* Look for SCCs */
  /* In the tight case, can't prune if there is only 1 SCC. */
  /* In the loose case, can't prune if there is only 0 SCC. */
  if (nunreach>0 && find_sccs(1) > (slack==0)) {  
    /* Delete edges */
    compress_edges(pdata->vartarget,nvartargets);
  }
  return TRUE;
} /* all_dist_filter */




static void all_dist_init MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                                 struct all_dist_data *pdata,
                                 TAGGED vals)
{
  VERTEX varvertex, valvertex;
  int i=0;
  FDITER it;
  TAGGED t;
  BOOL dense = TRUE;

  /* First, allocate all value vertices. */
  valvertex = pdata->nvars;
  fditer_init(&it, vals);
  while (!fditer_empty(&it)) {
    t = fditer_next(&it);
    VAL(valvertex) = t;
    MATE(valvertex) = EOL;
    TVAL(i++) = valvertex;
    valvertex++;
    if (GetSmall(t)!=i)
      dense = FALSE;
  }
  pdata->dense = dense;

  /* Then, allocate all var vertices and their edges */
  varvertex = 0;
  for (i=0; i<pdata->nvars; i++)
    {
      VAL(varvertex) = MakeSmall(i+1);
      MATE(varvertex) = EOL;
      TVAR(i) = varvertex;
      varvertex++;
    }
} /* all_dist_init */

static void ass_init MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                            struct all_dist_data *pdata,
                            TAGGED tmat)
{
  VERTEX vertex = 0;
  int k;
  int n = pdata->nvars;
  TAGGED trow, telt;

  pdata->dense = TRUE;

  for (k=0; k<n; k++) {
    VAL(vertex) = MakeSmall(k+1);
    MATE(vertex) = EOL;
    TVAR(k) = vertex;
    vertex++;
  }
  for (k=0; k<n; k++) {
    VAL(vertex) = MakeSmall(k+1);
    MATE(vertex) = EOL;
    TVAL(k) = vertex;
    vertex++;
  }
  if (!IsAtomic(tmat)) {
    k = 0;
    while (tmat!=atom_nil) {
      DerefCar(trow,tmat);
      DerefCdr(tmat,tmat);
      while (trow!=atom_nil) {
	DerefCar(telt,trow);
	DerefCdr(trow,trow);
	pdata->cost[k++] = GetSmall(telt);
      }
    }	  
  }			      
}

#define EVKEY(I) (event.key[I])

static TAGGED cons_batch MAGIC (HIDDEN_PROTO
				int *buf2, int n)
{
  TAGGED t1, t2;
  int n2;

  switch (n) {
  case 1:
    t1 = MakeSmall(buf2[0]+1);
    return fd_interval(t1,t1);
  case 2:
    t1 = MakeSmall(buf2[0]+1);
    t2 = MakeSmall(buf2[1]+1);
    return fd_pair(t1,t2);
  default:
    n2 = n>>1;
    return fd_or(cons_batch(buf2,n2),cons_batch(buf2+n2,n-n2));
  }
}

/*
  Pruning rules:
  [P1.] Xi in {j | i in dom(Yj)}
  [P2.] Yj in {i | j in dom(Xi)}

  A sweep algorithm:

  Events for rule P1:
  add(i,j) where i=min(I), for each interval I of Yj.
  del(i,j) where i=max(I), for each interval I of Yj.
  query(i), for each i.

  Sort events by (1) i, (2) add<query<del.

  Data structures: FD set.

  Handle add(_,j): add j to FD set.
  Handle del(_,j): delete j from FD set.
  Handle query(i): Infer Xi in FD set.
*/
static BOOL pairing_filter MAGIC (HIDDEN_PROTO
				  struct all_dist_data *pdata)
{
  /* volatile storage: */
  int *buf1, *buf2, *buf3, ap, dp, minv[2], maxv[2];
  int nxs = pdata->nvars;
  int ntargets = pdata->nvartargets;
  int nevents, iter, ev, type, i, k, elt, rvertex;
  Dvar dv;
  TAGGED feasible;
  struct {
    int *key;
    int *type;
    int *elt;
    int *queue;
  } event;

  /* At most ntargets QUERY events,
     at most ntargets*ceiling(nxs/2) ADD events,
     at most ntargets*ceiling(nxs/2) DEL events.
     Hence, at most ntargets*(nxs+2) events.
  */
  nevents = ntargets*(nxs+2);
  buf1 = Malloc(nxs,int);
  buf2 = Malloc(nevents,int);
  buf3 = Malloc(nevents,int);
  event.queue = Malloc(nevents,int);
  event.key = Malloc(nevents,int);
  event.type = Malloc(nevents,int);
  event.elt = Malloc(nevents,int);
  minv[0] = minv[1] = nxs;
  maxv[0] = maxv[1] = -1;
  for (iter=0; iter<2; iter++)
    for (i=0; i<ntargets; i++) {
      elt = iter ? TVAL(i)-nxs : TVAR(i);
      if (minv[iter]>elt) minv[iter] = elt;
      if (maxv[iter]<elt) maxv[iter] = elt;
    }
  for (iter=0; iter<2; iter++) {
    feasible = EmptySet;
    /* Use the low part of buf2 for buffered ADD events. */
    /* Use the high part of buf2 for buffered DEL events. */
    ap = 0;
    ev = 0;
    for (type=0; type<3; type++)
      for (k=0; k<2; k++)
	for (i=0; i<ntargets; i++) {
	  elt = k ? TVAL(i) : TVAR(i);
	  dv = DVAR(elt);
	  if (!dvar_is_integer_first(dv) &&
	      /* If the variable is ground _on entry_, we
		 rely on its being taken care of by pairing/2.
		 We know for a fact that pairing/2 is not
		 in the propagation queue. */
	      (iter ^ k ^ (type==1))) {
	    if (type==1) {
	      event.key[ev] = (!k ? elt : elt-nxs);
	      event.type[ev] = 1;
	      event.elt[ev] = elt;
	      event.queue[ev] = ev;
	      ev++;
	    } else if (pdata->incremental && LIVE(elt)==SIZE(elt)) {
	      if (type==0)
		buf2[ap++] = iter ? elt : elt-nxs;
	    } else {
	      DVITER it;
	      long min, max;
	      dviter_init(&it, dv);
	      while (!dviter_empty(&it)) {
		dviter_next_interval_l(&it, &min, &max);
		/* ADD event useful if interval overlaps min..max */
		/* DEL event useful if end point inside min..max */
		if (max-1>=minv[iter]) {
		  if (type==0 && min-1<=minv[iter])
		    buf2[ap++] = iter ? elt : elt-nxs;
		  else if (type==0 ? (min-1<=maxv[iter])
			   : (max-1< maxv[iter])) {
		    event.key[ev] = (type==0 ? min-1 : max-1);
		    event.type[ev] = type;
		    event.elt[ev] = elt;
		    event.queue[ev] = ev;
		    ev++;
		  }
		}
	      }
	    }
	  }
	}
    dp = nevents = ev;
    KEYSORT(event.queue, nevents, int,
	    buf1, nxs, EVKEY, buf3);
    for (i=0; i<nevents; i++) {
      event.queue[i] = ev = buf3[i];
    }
    for (i=0; i<nevents; i++) {
      ev = event.queue[i];
      elt = event.elt[ev];
      type = event.type[ev];
      switch (type) {
      case 0:			/* add */
	if (dp<nevents) {
	  feasible = fd_subtract(feasible,cons_batch(buf2+dp, nevents-dp));
	  dp = nevents;
	}
	buf2[ap++] = iter ? elt : elt-nxs;
	break;
      case 2:			/* del */
	if (ap>0) {
	  feasible = fd_or(feasible,cons_batch(buf2, ap));
	  ap = 0;
	}
	buf2[--dp] = iter ? elt : elt-nxs;
	break;
      default:			/* query */
	if (ap>0) {
	  feasible = fd_or(feasible,cons_batch(buf2, ap));
	  ap = 0;
	}
	if (dp<nevents) {
	  feasible = fd_subtract(feasible,cons_batch(buf2+dp, nevents-dp));
	  dp = nevents;
	}
	if (feasible==EmptySet)
	  goto ret;
	dv = DVAR(elt);
	switch (dvar_fix_set(dv,feasible)) {
	case -1:
	  goto ret;
	case 0:
	  break;
	default:
	  SIZE(elt) = dvar_value_count(dv);
	  pdata->pruned = TRUE;
	}
      }
    }
  }
  for (k=0; k<2; k++)
    for (i=0; i<ntargets; i++) {
      elt = k ? TVAL(i) : TVAR(i);
      if (!pdata->incremental || LIVE(elt)>SIZE(elt)) {
	rvertex = MATE(elt);
	if (rvertex>EOL && !dvar_contains_value_t(DVAR(elt),VAL(rvertex)))
	  MATE(elt) = MATE(rvertex) = EOL;
	LIVE(elt) = SIZE(elt);
      }
    }
 ret:
  SP_free(buf1);
  SP_free(buf2);
  SP_free(buf3);
  SP_free(event.queue);
  SP_free(event.key);
  SP_free(event.type);
  SP_free(event.elt);
  return (iter==2);
}

static BOOL ass_filter MAGIC (HIDDEN_PROTO
			      struct all_dist_data *pdata)
{
  int i;
  VERTEX vertex;
  int nvartargets = pdata->nvartargets;
  int nvaltargets = pdata->nvaltargets;
  
  for (i=0; i<nvartargets; i++) {
    vertex = TVAR(i);
    COMPONENT(vertex) = 0;
    VISITED(vertex) = 0;
  }
  for (i=0; i<nvaltargets; i++) {
    vertex = TVAL(i);
    COMPONENT(vertex) = 0;
    VISITED(vertex) = 0;
  }
  pdata->slack = 0;
  /* Propagate ground values until fixpoint.  NOT OPTIONAL,
     the SCC search skips singletons. */
  if (!propagate_value(nvartargets,nvaltargets))
    return FALSE;
  /* Find maximal matching. */
  if (!augment(nvartargets))
    return FALSE;
  
  /* Look for SCCs */
  /* In the tight case, can't prune if there is only 1 SCC. */
  /* In the loose case, can't prune if there is only 0 SCC. */
  if (find_sccs(1) > 1) {  
    /* Delete edges */
    compress_edges(pdata->vartarget,nvartargets);
    compress_edges(pdata->valtarget,nvaltargets);
  }
  return TRUE;
} /* ass_filter */

/* A feasibility check: a necessary condition for circuit/[1,2]
   is that the graph be a single, strongly connected component.
   Implementation: an iterative version of Tarjan's algorithm + 
   a necessary condition for Hamiltonicity.
   */
static BOOL single_circuit MAGIC (HIDDEN_PROTO
				  struct all_dist_data *pdata,
				  VERTEX vertex,
				  VERTEX *tail)
{
  int id=0, top=0, min, m, vertid;
  int majormin=0, majormax=0;	/* first (last) id of latest major subtree */
  BOOL seen_major = TRUE;	/* seen edge to latest major subtree? */
  VERTEX child;
  VERTEX *vstack = pdata->dist_stack;
  long *lstack = (long *)vstack;
  TAGGED *tstack = (TAGGED *)vstack;
  int nvars = pdata->nvars;
  TAGGED tv;
  DVITER it;

start:
  VISITED(vertex) = vertid = min = ++id;
  dviter_init(&it, DVAR(vertex));
  while (!dviter_empty(&it)) {
    tv = dviter_next_value_t(&it);
    child = tail[neighbor_of(tv,vertex)-nvars /*VAL(neighs[j])-1*/];
    m = VISITED(child);
    if (m==0) {
      vstack[top] = vertex;
      tstack[top+1] = tv;
      lstack[top+2] = min;
      top += 3;
      vertex = child;
      goto start;
    } else if (vertid>1 && m>=majormin && m<=majormax)
      seen_major = TRUE;
  cont:
    if (min > m)
      min = m;
  }
  if (min==vertid)
    return (min==1 && id==pdata->nvartargets);
  m = min;
  top -= 3;
  vertex = vstack[top];
  tv = tstack[top+1];
  min = lstack[top+2];
				/* here, m is the minimal ID reachable from
				   some just visited child of vertex */
  if (VISITED(vertex)==1) {	/* just visited a major subtree */
    if (!seen_major && tv!=dvar_min_t(DVAR(vertex))) /* which must have an edge into the subtree */
      return FALSE;		/* to its left */
    majormin = vertid;
    majormax = id;
    seen_major = FALSE;
  }
  vertid = VISITED(vertex);
  dviter_init(&it, DVAR(vertex));
  dviter_skip_t(&it, tv);
  goto cont;
}

static void all_dist_alloc_temp MAGIC (HIDDEN_PROTO
				       struct all_dist_data *pdata)
{
  int nvars = pdata->nvars;
  int nvals = pdata->nvals;
  char *ptr = SP_malloc((4*nvars+2*nvals)*sizeof(VERTEX));
  
  pdata->dist_stack = (VERTEX *)ptr; /* volatile */
  ptr += (3*nvars+nvals)*sizeof(VERTEX);
  pdata->vertex.size = (long *)ptr; /* volatile */
  ptr += (nvars+nvals)*sizeof(VERTEX);
}

static void all_dist_free_temp MAGIC (HIDDEN_PROTO
				      struct all_dist_data *pdata)
{
  SP_free(pdata->dist_stack);
}

static struct all_dist_data *
all_dist_alloc MAGIC (HIDDEN_PROTO
		      Argdecl,
		      int nvars,int nvals,int nrefs, 
		      TAGGED handle,
		      int flags) /* 0x1 = circuit, 0x2 = cost matrix */
{
  long total_size = 
    nrefs*sizeof(struct dvar) +
    7*nvars*sizeof(TAGGED) +
#if WITNESS
    2*nvars*sizeof(TAGGED) +
#endif
#if INCTREE
    nvars*sizeof(TAGGED) +
#endif
    7*nvals*sizeof(TAGGED) +
    (!(flags & 0x2) ? 0 :
     3*(nvars*nvars)*sizeof(long) +
     4*nvars*sizeof(long) +
     8*nvars*sizeof(int));
  struct all_dist_data *pdata =
    Palloc(struct all_dist_data, total_size, handle);
  
  char *ptr;
  int i;
    
    pdata->destructor = all_distinct_destructor;
    FD_STORE_SPENV(pdata->spenv);
    pdata->refbase = SP_alloc_globrefs(2*nrefs);
    pdata->nrefs = 2*nrefs;
    pdata->nvars = nvars;
    pdata->nvartargets = nvars;
#if CSTATE_ONLY
    pdata->nvars_back = nvars;
#endif
    pdata->nvals = nvals;
    pdata->nvaltargets = nvals;
#if CSTATE_ONLY
    pdata->nvals_back = nvals;
#endif
    ptr = (char *)(pdata+1);
    pdata->dvar = (Dvar)ptr;
    ptr += nrefs*sizeof(struct dvar);
    pdata->vartarget = (VERTEX *)ptr; /* array + aux */
    ptr += 2*nvars*sizeof(VERTEX);
    pdata->valtarget = (VERTEX *)ptr; /* array + aux */
    ptr += 2*nvals*sizeof(VERTEX);
#if WITNESS
    pdata->sccwitix = (long *)ptr;
    ptr += nvars*sizeof(long);
    pdata->sccwit = (VERTEX *)ptr;
    ptr += nvars*sizeof(VERTEX);
#endif
#if INCTREE
    pdata->reachset = (VERTEX *)ptr;
    ptr += nvars*sizeof(VERTEX);
#endif
    pdata->vertex.val = (TAGGED *)ptr;
    ptr += (nvars+nvals)*sizeof(VERTEX);
    pdata->vertex.mate = (VERTEX *)ptr;
    ptr += (nvars+nvals)*sizeof(VERTEX);
    pdata->vertex.live = (long *)ptr;
    ptr += (nvars+nvals)*sizeof(VERTEX);
    pdata->vertex.component = (long *)ptr;
    ptr += (nvars+nvals)*sizeof(VERTEX);
    pdata->vertex.visited = (long *)ptr;
    ptr += (nvars+nvals)*sizeof(VERTEX);
    if (!(flags & 0x2)) {
      pdata->cost = NULL;
    } else {
      pdata->cost = (long *)ptr;
      ptr += nvars*nvars*sizeof(long);
      for (i=0; i<2; i++) {
	pdata->rc.rcost[i] = (long *)ptr;
	ptr += nvars*nvars*sizeof(long);
	pdata->rc.lbdual[i] = (long *)ptr;
	ptr += nvars*sizeof(long);
	pdata->rc.ubdual[i] = (long *)ptr;
	ptr += nvars*sizeof(long);
      }
      pdata->rc.index_var = (int *)ptr;
      ptr += 2*nvars*sizeof(int);
      pdata->rc.var_index = (int *)ptr;
      ptr += 2*nvars*sizeof(int);
      for (i=0; i<2; i++) {
	pdata->rc.lbprimal[i] = (int *)ptr;
	ptr += nvars*sizeof(int);
	pdata->rc.ubprimal[i] = (int *)ptr;
	ptr += nvars*sizeof(int);
      }
    }
#if DBG
    if (ptr != (char *)(pdata+1)+total_size)
      printf("SHOULDN'T HAPPEN: expected ptr=%p, got %p\n",
	     (char *)(pdata+1)+total_size, ptr);
#endif
    return pdata;
}


/* the targets need to be ordered (see ORDER comments) */
static void contract_vars (struct all_dist_data *pdata)
{
  int imax = pdata->nvartargets;
  int n = pdata->nvars;
  VERTEX *target = pdata->vartarget;
  int i, j, k;
    
  /* separate into nonground (0..) and ground (n..) */
  for (i=0, j=0, k=n; i<imax; i++) {
    VERTEX current = target[i];
    
    if (LIVE(current)>1)
      target[j++] = current;
    else {
      VISITED(current) = EOL;	/* maintain invariant! */
      target[k++] = current;
    }
  }
  if (i==j) return;
  
  pdata->nvartargets = j;

  while (n<k)
    target[j++] = target[n++];
}


/* the targets need to be ordered (see ORDER comments) */
static void contract_vals (struct all_dist_data *pdata)
{
  int imax = pdata->nvaltargets;
  int n = pdata->nvals;
  VERTEX *target = pdata->valtarget;
  int i, j, k;
    
  /* separate into nonground (0..) and ground (n..) */
  for (i=0, j=0, k=n; i<imax; i++) {
    VERTEX current = target[i];
    
    if (LIVE(current)>1)
      target[j++] = current;
    else {
      VISITED(current) = EOL;	/* maintain invariant! */
      target[k++] = current;
    }
  }
  if (i==j) return;
  
  pdata->nvaltargets = j;

  while (n<k)
    target[j++] = target[n++];
}

#define ASSIGNMENT_ARITY 8

/* '$fd_all_distinct'(+State0, -State, -Actions) :-
   State0 is f(LVec,Handle,Stamp) where LVec is a list of Var-Attribute
   State  similarly,
   Actions is a list of prunings etc.
   */
void SPCDECL 
prolog_fd_all_distinct MAGIC (HIDDEN_PROTO
			      SP_term_ref State0,
			      SP_term_ref State,
			      SP_term_ref Actions)
{
  WAMENV;
  int nvars, nvals;
  int i;
  TAGGED tlvec, telt, t1;
  TAGGED all, handle;
  long state_stamp;
  int ent = -1;			/* disentailed unless otherwise */
  struct all_dist_data *pdata;
  BOOL committed;
  BOOL ground = TRUE;

  (void)State0;                 /* [PM] 3.9b5 avoid -Wunused */

  
  w->numstack_end = NULL;

/*    X(0) = RefTerm(State0); */
  dvar_export_start(3);
  RefTerm(State) = unify_output_state(w,&handle,&state_stamp,&committed);

  if (!IsVar(handle)) {		/* got [Flag | '$free'(Ptr)] */
    fd.gdata = pdata = Pdata(struct all_dist_data,handle);
    nvars = pdata->nvars;
    nvals = pdata->nvals;
  } else {			/* build persistent state */
				/* compute nvars, nvals, all */
    all = EmptySet;
    nvars = 0;
    DerefArg(tlvec,X(0),1);
    while (TagIsLST(tlvec)) {
      DerefCar(telt,tlvec);
      DerefCdr(tlvec,tlvec);
      DerefArg(t1,telt,2);	/* Attribute */
      DerefAttribute(t1,t1);
      all = fd_merge_into(DomainSet(t1),all);
      nvars++;
    }
    if (nvars==0) {
      ent = 1;
      goto ret1;
    }
    nvals = fd_size(all);
    fd.gdata = pdata = all_dist_alloc(w,nvars,nvals,nvars,handle,0x0);
    pdata->stamp = state_stamp-1; /* DON'T trust initially */
    DerefArg(tlvec,X(0),1);
    for (i=0; i<nvars; i++)
      {
	DerefCar(telt,tlvec);
	DerefCdr(tlvec,tlvec);
	DerefArg(t1,telt,1);	/* FD variable */
	RefGlob(RefVar(i)) = t1;	/* protect it */
	DerefArg(t1,telt,2);	/* attribute */
	RefGlob(RefAttr(i)) = t1;	/* protect it */
      }
    all_dist_init(pdata,all);
  }

				/* RESUME HERE */
  all_dist_alloc_temp(pdata);
  pdata->pruned = FALSE;
  pdata->incremental = (state_stamp==pdata->stamp);
  if (!pdata->incremental) {
    int nvars1, nvals1;
#if CSTATE_ONLY
    nvars1 = pdata->nvars_back;
    nvals1 = pdata->nvals_back;
#else
    DerefArg(telt,X(0),2);
    nvars1 = nvars-GetSmall(telt);
    DerefArg(telt,X(0),3);
    nvals1 = nvals-GetSmall(telt);
#endif
    if (pdata->nvartargets < nvars1) {
      pdata->nvartargets = nvars1;
      qsort_asc_val(pdata->vartarget,nvars1);
    }
    if (pdata->nvaltargets < nvals1) {
      pdata->nvaltargets = nvals1;
      qsort_asc_val(pdata->valtarget,nvals1);
    }
  }

  /* Always refresh dom fields */
  for (i=0; i<pdata->nvartargets; i++) {
    VERTEX vertex = TVAR(i);
    VERTEX mate = MATE(vertex);
    
    dvar_init(DVAR(vertex), RefAttr(vertex), RefVar(vertex));
    t1 = RefGlob(RefAttr(vertex)); /* get attribute */
    DerefAttribute(t1,t1);
    LIVE(vertex) = SIZE(vertex) = GetSmall(DomainSize(t1));
    if (LIVE(vertex)>1) ground = FALSE;
    VISITED(vertex) = 0;
    if (mate>EOL && !dvar_contains_value_t(DVAR(vertex),VAL(mate)))
      MATE(vertex) = MATE(mate) = EOL;
  }
  pdata->stamp = state_stamp+1;
  if (!all_dist_filter(pdata))
    goto ret;
  ent = ground;
  if (!pdata->pruned)
    goto ret;

				/* Compute prunings. */
  for (i=0; i<pdata->nvartargets; i++) {
    VERTEX vertex = TVAR(i);
    
    dvar_pruning_done(DVAR(vertex));
  }
  for (i=0; i<pdata->nvartargets; i++) {
    VERTEX vertex = TVAR(i);
    
    dvar_export(DVAR(vertex), 3);
  }
  contract_vars(pdata);
  for (i=0; i<pdata->nvaltargets; i++) {
    VERTEX vertex = TVAL(i);
    VERTEX mate = MATE(vertex);

    LIVE(vertex) = mate>EOL ? LIVE(mate) : pdata->slack ? 2 : 0;
  }
  contract_vals(pdata);

#if CSTATE_ONLY
  if (committed) {
    pdata->nvars_back = pdata->nvartargets;
    pdata->nvals_back = pdata->nvaltargets;
  }
#else
  CTagToArg(X(0),2) = MakeSmall(nvars-pdata->nvartargets);
  CTagToArg(X(0),3) = MakeSmall(nvals-pdata->nvaltargets);
#endif
  
  ent = (pdata->nvartargets<=1);
#if DBG > 1
  if (ent>=0) {			/* check invariants */
    printf("%% invariants\n");
    for (i=0; i<pdata->nvaltargets; i++) {
      VERTEX vertex = TVAL(i);
      VERTEX mate = MATE(vertex);

      if (mate>EOL && LIVE(mate)==1)
	printf("! singleton var=%d matched non-singleton val=%d\n", (int)mate, (int)vertex);
    }
  }
#endif
 ret:
  all_dist_free_temp(pdata);
  Pfree(ent==1,handle,pdata);
 ret1:
  dvar_export_done(Actions, ent, 3);
}



/* '$fd_assignment'(+State0, -State, -Actions) :-
   State0 is f(LVec,RVec,Handle,Stamp) where LVec,RVec are lists of f(Var,DomainMut)
   State  similarly,
   Actions is a list of prunings etc.
   */
void SPCDECL 
prolog_fd_assignment MAGIC (HIDDEN_PROTO
			    SP_term_ref State0,
			    SP_term_ref State,
			    SP_term_ref Actions)
{
  WAMENV;
  int nvars, nvals, nrefs;
  int i, j, flags;
  TAGGED tlvec, telt, trvec, t1;
  TAGGED handle;
  long state_stamp;
  int ent = -1;			/* disentailed unless otherwise */
  struct all_dist_data *pdata;
  BOOL committed, ground=TRUE;

  (void)State0;                 /* [PM] 3.9b5 avoid -Wunused */

  
  w->numstack_end = NULL;

/*    X(0) = RefTerm(State0); */
  dvar_export_start(3);
  RefTerm(State) = unify_output_state(w,&handle,&state_stamp,&committed);
  DerefArg(telt,X(0),4);
  flags = GetSmall(telt);	/* 0x1=circuit, 0x2=cost */

  if (!IsVar(handle)) {		/* got [Flag | '$free'(Ptr)] */
    fd.gdata = pdata = Pdata(struct all_dist_data,handle);
    nvars = pdata->nvars;
    nvals = pdata->nvals;
  } else {			/* build persistent state */
				/* compute nvars, nvals */
    nvars = 0;
    DerefArg(tlvec,X(0),1);
    while (TagIsLST(tlvec)) {
      DerefCdr(tlvec,tlvec);
      nvars++;
    }
    if (nvars==0) {
      ent = 1;
      goto ret1;
    }
    nvals = nvars;
    DerefArg(telt,X(0),5);
    nrefs = IsAtomic(telt) ? 2*nvars : 2*nvars+1;
    fd.gdata = pdata = all_dist_alloc(w,nvars,nvals,nrefs,handle,flags);
    pdata->stamp = state_stamp-1; /* DON'T trust initially */
    DerefArg(tlvec,X(0),1);
    DerefArg(trvec,X(0),2);
    for (i=0; i<nvars; i++)
      {
	DerefCar(telt,tlvec);
	DerefCdr(tlvec,tlvec);
	DerefArg(t1,telt,1);	/* FD variable */
	RefGlob(RefVar(i)) = t1;	/* protect it */
	DerefArg(t1,telt,2);	/* attribute */
	RefGlob(RefAttr(i)) = t1;	/* protect it */
	DerefCar(telt,trvec);
	DerefCdr(trvec,trvec);
	DerefArg(t1,telt,1);	/* FD variable */
	RefGlob(RefVar(nvars+i)) = t1;	/* protect it */
	DerefArg(t1,telt,2);	/* attribute */
	RefGlob(RefAttr(nvars+i)) = t1;	/* protect it */
      }
    DerefArg(telt,X(0),5);
    if (!IsAtomic(telt)) {
      DerefArg(t1,telt,1);	/* cost variable */
      RefGlob(RefVar(nrefs-1)) = t1;
      DerefArg(t1,telt,2);	/* cost variable attribute */
      RefGlob(RefAttr(nrefs-1)) = t1;
      pdata->rc.costvar = pdata->dvar+(nrefs-1);
      DerefArg(telt,telt,3);	/* cost matrix */
    }
    ass_init(pdata,telt);
  }

				/* RESUME HERE */
  all_dist_alloc_temp(pdata);
  pdata->pruned = FALSE;
  pdata->incremental = (state_stamp==pdata->stamp);
  if (!pdata->incremental) {
    int nvars1, nvals1;
#if CSTATE_ONLY
    nvars1 = pdata->nvars_back;
    nvals1 = pdata->nvals_back;
#else
    DerefArg(telt,X(0),3);
    nvars1 = nvars-GetSmall(telt);
    nvals1 = nvals-GetSmall(telt);
#endif
    if (pdata->nvartargets < nvars1) {
      pdata->nvartargets = nvars1;
      qsort_asc_val(pdata->vartarget,nvars1);
    }
    if (pdata->nvaltargets < nvals1) {
      pdata->nvaltargets = nvals1;
      qsort_asc_val(pdata->valtarget,nvals1);
    }
  }
  
  /* Always refresh dom fields */
  for (i=0; i<pdata->nvartargets; i++) {
    VERTEX vertex = TVAR(i);

    dvar_init(DVAR(vertex), RefAttr(vertex), RefVar(vertex));
    t1 = RefGlob(RefAttr(vertex)); /* get attribute */
    DerefAttribute(t1,t1);
    SIZE(vertex) = GetSmall(DomainSize(t1));
    if (SIZE(vertex)>1) ground = FALSE;
    VISITED(vertex) = 0;
  }
  for (i=0; i<pdata->nvaltargets; i++) {
    VERTEX vertex = TVAL(i);

    dvar_init(DVAR(vertex), RefAttr(vertex), RefVar(vertex));
    t1 = RefGlob(RefAttr(vertex)); /* get attribute */
    DerefAttribute(t1,t1);
    SIZE(vertex) = GetSmall(DomainSize(t1));
    if (SIZE(vertex)>1) ground = FALSE;
    VISITED(vertex) = 0;
  }
  pdata->stamp = state_stamp+1;
  if (!pairing_filter(pdata))
    goto ret;
  if (!pdata->cost ? !ass_filter(pdata) :
      !cost_filter(pdata,GetSmall(CTagToArg(X(0),6)),flags))
    goto ret;
  /* If <6 targets and more than one SCC, then one
     will be of size 2 and will be caught by pairing/2. */
  if ((flags&1) && !pdata->cost && pdata->nvartargets>=6) {
    VERTEX v1, v2, *head, *tail;
    VERTEX bestv = EOL;
    int bestval = 0;

    head = Malloc(nvars,VERTEX);
    tail = Malloc(nvars,VERTEX);
    for (i=0; i<nvars; i++)
      head[i] = tail[i] = i;
    for (i=0; i<pdata->nvartargets; i++) {
      v1 = TVAR(i);
      VISITED(v1) = 0;
      if (LIVE(v1) > bestval) {
	bestv = v1;
	bestval = LIVE(v1);
      }
    }
    for (; i<nvars; i++) {
      v1 = TVAR(i);
      v2 = dvar_min_l(DVAR(v1))-1;
      v1 = head[v1];
      v2 = tail[v2];
      tail[v1] = v2;
      head[v2] = v1;
    }
    if (!single_circuit(pdata,bestv,tail))
      bestv = EOL;
    SP_free(head);
    SP_free(tail);
    if (bestv==EOL)
      goto ret;
  }
  ent = ground;
  if (!pdata->pruned)
    goto ret;
				/* Compute prunings */
  for (i=0; i<pdata->nvartargets; i++)
    for (j=0; j<2; j++) {
      VERTEX vertex = j==0 ? TVAR(i) : TVAL(i);

      dvar_pruning_done(DVAR(vertex));
    }
  if (pdata->cost)
    dvar_pruning_done(pdata->rc.costvar);
  for (i=0; i<pdata->nvartargets; i++)
    for (j=0; j<2; j++) {
      VERTEX vertex = j==0 ? TVAR(i) : TVAL(i);

      dvar_export(DVAR(vertex), 3);
    }
  if (pdata->cost) {
    long costbase = GetSmall(CTagToArg(X(0),6));
    
    dvar_export(pdata->rc.costvar,3);
    for (i=0; i<pdata->nvartargets; i++) {
      VERTEX vertex = TVAR(i);
      if (LIVE(vertex)==1)
	costbase += pdata->cost[nvars*vertex + dvar_min_l(DVAR(vertex))-1];
      CTagToArg(X(0),6) = MakeSmall(costbase);
    }
  }

  contract_vars(pdata);
  contract_vals(pdata);
#if DBG
  if (pdata->nvartargets!=pdata->nvaltargets)
    printf("SHOULDN'T HAPPEN: nvartargets=%d, nvaltargets=%d\n",
	   pdata->nvartargets, pdata->nvaltargets);
#endif

#if CSTATE_ONLY
  if (committed) {
    pdata->nvars_back = pdata->nvartargets;
    pdata->nvals_back = pdata->nvaltargets;
  }
#else
  CTagToArg(X(0),3) = MakeSmall(nvars-pdata->nvartargets);
#endif

  ent = (pdata->nvartargets<=1);
 ret:
  all_dist_free_temp(pdata);
  Pfree(ent==1,handle,pdata);
 ret1:
  dvar_export_done(Actions, ent, 3);
  if (ent==1)
    CTagToArg(X(0),ASSIGNMENT_ARITY-1) = TaggedZero; /* hide state from helper */
}



/*** COST-BASED PRUNING ***/

/* iterating over a "row" (in terms of column indices) */

#define rowiter DVITER

#define rowiter_init(IT,R) dviter_init(IT,DVAR(pdata->rc.index_var[R]))

#define rowiter_empty(IT) dviter_empty(IT)

#define rowiter_next(IT) (pdata->rc.var_index[dviter_next_value_l(IT)+nvars-1]-n)

/* iterating over a "column" (in terms of row indices) */

#define coliter DVITER

#define coliter_init(IT,R) dviter_init(IT,DVAR(pdata->rc.index_var[(R)+n]))

#define coliter_empty(IT) dviter_empty(IT)

#define coliter_next(IT) (pdata->rc.var_index[dviter_next_value_l(IT)-1])

static int cmp_asc_long(long *l1, long *l2)
{
  long val1 = *l1;
  long val2 = *l2;

  return CMP(val1,val2);
}

#define QType long
#define QCmp  cmp_asc_long
#define QSort qsort_asc_long
#include "qsort.ic"

#if 0 /* MAY COME IN HANDY */
static int cmp_asc_cost MAGIC (HIDDEN_PROTO
			       int *t1, int *t2)
{
  struct all_dist_data *pdata = fd.gdata;
  long c1, c2;
  
  c1 = -pdata->rc.cost[*t1];
  c2 = -pdata->rc.cost[*t2];
  if (c1==c2)
    return 0;
  else
    return (c1<c2 ? -1 : 1);
}

#define QType int
#define QCmp  cmp_asc_cost
#define QSort qsort_asc_cost
#include "qsort.ic"
#endif

/* BINARY HEAPS */

#define SWAP(I,J)				\
{						\
  int vi = pdata->rc.heap[I];			\
  int vj = pdata->rc.heap[J];			\
  pdata->rc.heap[I] = vj;			\
  pdata->rc.heap[J] = vi;			\
  pdata->rc.vheap[vi] = (J);			\
  pdata->rc.vheap[vj] = (I);			\
}

static void decrease_key(struct all_dist_data *pdata,int v, long value)
{
  int i = pdata->rc.vheap[v];
  int p = (i-1)>>1;
  pdata->rc.distrow[v] = value;
  while (i>0 && value<pdata->rc.distrow[pdata->rc.heap[p]]) {
    SWAP(i,p);
    i = p;
    p = (i-1)>>1;
  }
}

static void spheapify(struct all_dist_data *pdata,int i)
{
  int *heap = pdata->rc.heap;
  long *distrow = pdata->rc.distrow;
  
  for (;;) {
    int l = (i<<1)+1;
    int smallest = i;
    if (l<pdata->rc.heapsize && distrow[heap[l]]<distrow[heap[smallest]])
      smallest = l;
    if (l+1<pdata->rc.heapsize && distrow[heap[l+1]]<distrow[heap[smallest]])
      smallest = l+1;
    if (smallest==i)
      break;
    SWAP(i,smallest);
    i = smallest;
  }
}

static void spheap_init(struct all_dist_data *pdata,int k)
{
  int n = pdata->rc.ntargets;
  int i, j;
  for (i=0; i<n-1; i++)
    pdata->rc.heap[i] = pdata->rc.vheap[i] = i;
  pdata->rc.heap[k] = i;
  pdata->rc.vheap[i] = k;
  pdata->rc.heapsize = i;
  for (j=(i-2)>>1; j>=0; j--)
    spheapify(pdata,j);
}

static int spheap_extract_min(struct all_dist_data *pdata)
{
  int u = pdata->rc.heap[0];
  int v = pdata->rc.heap[--pdata->rc.heapsize];

  if (pdata->rc.heapsize>0) {
    pdata->rc.heap[0] = v;
    pdata->rc.vheap[v] = 0;
    spheapify(pdata,0);
  }
  return u;
}

static void all_pairs_shortest_paths MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                                     struct all_dist_data *pdata,
				     int **primal)
{
  int n = pdata->rc.ntargets;
  int k;
  for (k=0; k<n; k++) {
    /* solve single-source shortest-paths for vertex k */
    /* warm start: exclude k from queue */
    pdata->rc.distrow = pdata->rc.dist + n*k;
    spheap_init(pdata,k);
    while (pdata->rc.heapsize>0) {
      int u, v;
      long du, duv;
      coliter it;
      u = spheap_extract_min(pdata);
      du = pdata->rc.distrow[u];
      if (du<CLPFD_MAXINT) {
	coliter_init(&it,primal[0][u]);
	while (!coliter_empty(&it)) {
	  v = coliter_next(&it);
	  duv = pdata->rc.dist[n*u + v];
	  if (u!=v && pdata->rc.distrow[v]>du+duv)
	    decrease_key(pdata,v,du+duv);
	}
      }
    }
  }
}

#define safeadd(S,T) 				\
((S)==CLPFD_MAXINT ? CLPFD_MAXINT 		\
: (T)==CLPFD_MAXINT ? CLPFD_MAXINT 		\
: (S)+(T)					\
)

#define reduced_cost(I,J) 			\
(cost[n*(I)+(J)]-(dual[0][I]+dual[1][J]))

#define admissible(I,J) 			\
(reduced_cost(I,J)==0)

static int hungarian_init MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                                struct all_dist_data *pdata,
                                long **dual, int **primal)
{
  int i, j;
  rowiter it;
  int card = 0;
  int n = pdata->rc.ntargets;
  int nvars = pdata->nvars;
  long *cost = pdata->rc.cost;
  long lu, lv;
  
  /* identify a dual feasible solution */
  for (i=0; i<n; i++) {
    lu = CLPFD_MAXINT;
    rowiter_init(&it,i);
    while (!rowiter_empty(&it)) {
      j = rowiter_next(&it);
      lv = cost[n*i + j];
      if (lu>lv) lu = lv;
    }
    dual[0][i] = lu;
  }
  for (j=0; j<n; j++) {
    lu = CLPFD_MAXINT;
    coliter_init(&it,j);
    while (!coliter_empty(&it)) {
      i = coliter_next(&it);
      lv = cost[n*i + j] - dual[0][i];
      if (lu>lv) lu = lv;
    }
    dual[1][j] = lu;
  }
  /* find an initial allocation */
  for (j=0; j<n; j++)
    primal[1][j] = -1;		/* mark all columns unallocated */
  for (i=0; i<n; i++) {
    primal[0][i] = -1;		/* mark row i unallocated */
    rowiter_init(&it,i);
    while (!rowiter_empty(&it)) {
      j = rowiter_next(&it);
      if (primal[1][j] == -1 && admissible(i,j)) {
	primal[0][i] = j;	/* mark row i allocated */
	primal[1][j] = i;	/* mark col j allocated */
	card++;
	break;
      }
    }
  }
  return card;
}



static BOOL
hungarian_body MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                      struct all_dist_data *pdata,
                      long **dual, int **primal, int card, long *opt)
{
  int n = pdata->rc.ntargets;
  int nvars = pdata->nvars;
  int i, j, head, tail;
  long lu, lv;
  rowiter it;
  long *cost = pdata->rc.cost;

  /* labeling for checking max flow */
  while (card<n) {
    for (j=0; j<n; j++)
      pdata->rc.label[1][j] = -1;
    for (i=0; i<n; i++)
      pdata->rc.label[0][i] =  (primal[0][i] == -1) ? n : -1;
    /* for each col j, find index of labeled row in which min occurs */
    for (j=0; j<n; j++) {
      int ix = -1;
      lu = CLPFD_MAXINT;
      coliter_init(&it,j);
      while (!coliter_empty(&it)) {
	i = coliter_next(&it);
	if (pdata->rc.label[0][i] != -1) {
	  lv = reduced_cost(i,j);
	  if (lu>=lv)
	    lu = lv, ix = i;
	}
      }
      pdata->rc.index[j] = ix;
    }
  mklists:
    head = tail = 0;
    for (i=0; i<n; i++)
      if (pdata->rc.label[0][i] != -1)
	pdata->rc.queue[tail++] = i; /* enqueue row i */
    while (head!=tail) {
      i = pdata->rc.queue[head++];
      if (i<n) {		/* scan labeled row i */
	rowiter_init(&it,i);
	while (!rowiter_empty(&it)) {
	  j = rowiter_next(&it);
	  if (pdata->rc.label[1][j] == -1 && admissible(i,j)) { /* (i,j) admissible, j unlabeled */
	    pdata->rc.label[1][j] = i;
	    pdata->rc.queue[tail++] = j+n;	/* enqueue col j */
	  }
	}
      } else {
	j = i-n;		/* scan labeled col j */
	i = primal[1][j];
	if (i == -1) {		/* BREAKTHRU: found an augmenting path */
	  int row, col;

	  col = j;
	  for (;;) {
	    row = pdata->rc.label[1][col];
	    col = pdata->rc.label[0][row];
	    if (col==n) break;
	    primal[0][row] = primal[1][col] = -1; /* unalloc (row,col) */
	  }
	  col = j;
	  while (col<n) {
	    row = pdata->rc.label[1][col];
	    primal[0][row] = col; /* alloc (row,col) */
	    primal[1][col] = row;
	    col = pdata->rc.label[0][row];
	  }
	  card++;
	  goto augment;
	}
	if (pdata->rc.label[0][i] == -1) { /* (i,j) is alloced, i unlabeled */
	  pdata->rc.label[0][i] = j;
	  pdata->rc.queue[tail++] = i;	/* enqueue row i */
	  /* update indices for unlabeled cols. */
	  rowiter_init(&it,i);
	  while (!rowiter_empty(&it)) {
	    j = rowiter_next(&it);
	    if (pdata->rc.label[1][j] == -1) {
	      int ix = pdata->rc.index[j];
	      lv = reduced_cost(i,j);
	      if (lv==0) {
		pdata->rc.label[1][j] = i;
		pdata->rc.queue[tail++] = j+n; /* enqueue col j */
	      } else if (ix == -1 || reduced_cost(ix,j)>lv)
		pdata->rc.index[j] = i;
	    }
	  }
	}
      }
    }
    /* NONBREAKTHROUGH */
    /* find min. reduced cost in (LABELED,UNLABELED) */
    lu = CLPFD_MAXINT;
    for (j=0; j<n; j++)
      if (pdata->rc.label[1][j] == -1) {
	int ix = pdata->rc.index[j];
	if (ix != -1) {
	  lv = reduced_cost(ix,j);
	  if (lu>lv) lu = lv;
	}
      }
    if (lu==CLPFD_MAXINT)
      return FALSE;			/* no feasible assignment */
    /* update dual soln */
    for (i=0; i<n; i++)
      if (pdata->rc.label[0][i] != -1)
	dual[0][i] += lu;
    for (j=0; j<n; j++)
      if (pdata->rc.label[1][j] != -1)
	dual[1][j] -= lu;
    goto mklists;
  augment:
    ;
  }
  /* have max. cardinality matching */
  lu = 0;
  for (i=0; i<n; i++)
    lu += dual[0][i];
  for (j=0; j<n; j++)
    lu += dual[1][j];
  *opt = lu;
  lu = 0;
  return TRUE;
}

static BOOL
cost_based_filter MAGIC (DYMA_HIDDEN_PROTO /* [PM] 3.10.1 DynamicMallocBase */
                         struct all_dist_data *pdata, BOOL b /* 0=min, 1=max */)
{
  int n = pdata->rc.ntargets;
  int nvars = pdata->nvars;
  int i, j, nextj, u, v, card;
  long max_regret, min_regret;
  long **dual = (b==0 ? pdata->rc.lbdual : pdata->rc.ubdual);
  int  **primal = (b==0 ? pdata->rc.lbprimal : pdata->rc.ubprimal);
  long *cost = pdata->rc.cost;
  long *rcost = pdata->rc.rcost[b];
  
  for (i=0; i<n; i++) {
    rowiter it;			/* iterator for one variable */
    u = pdata->rc.index_var[i];
    rowiter_init(&it,i);
    nextj = rowiter_next(&it);
    for (j=0; j<n; j++) {
      v = pdata->rc.index_var[j+n]-nvars;
      if (j<nextj)
	pdata->rc.cost[n*i + j] = CLPFD_MAXINT;
      else if (b==0) {
	pdata->rc.cost[n*i + j] =  pdata->cost[nvars*u + v];
	nextj = rowiter_empty(&it) ? n : rowiter_next(&it);
      } else {
	pdata->rc.cost[n*i + j] = -pdata->cost[nvars*u + v];
	nextj = rowiter_empty(&it) ? n : rowiter_next(&it);
      }
    }
  }
  if (pdata->incremental) {
    card = n;
    for (i=0; i<n; i++) {
      j = primal[0][i];
      if (!dvar_contains_value_t(DVAR(pdata->rc.index_var[i]),
				 VAL(pdata->rc.index_var[j+n]))) {
	primal[0][i] = -1;
	primal[1][j] = -1;
	card--;
      }
    }
  } else {
    card = hungarian_init(pdata,dual,primal);
  }
  if (!hungarian_body(pdata,dual,primal,card,&pdata->rc.optcost[b]))
    return FALSE;
  /* We have obtained:
     pdata->rc.optcost, the (smallest) cost of an (optimal) assignment.

     PROPOSITION: all solutions containing only (xi=vi) with regret=0
     have cost = cost of optimal assignment.  If true, the following
     rule is valid:

     If optcost is NOT in the domain of costvar, then infer:
         min(costvar) >= optcost + smallest regret>0.

     Open question: how to infer a sharp bound for costvar.
  */
  if (b==0) {
    TAGGED ubt = dvar_max_t(pdata->rc.costvar);
    
    pdata->rc.optcost[0] += pdata->rc.costbase;
    pdata->rc.threshold[0] =
      TagIsSmall(ubt) ? GetSmall(ubt) - pdata->rc.optcost[0] :
      CLPFD_MAXINT-1;
  } else {
    TAGGED lbt = dvar_min_t(pdata->rc.costvar);
    
    pdata->rc.optcost[1] = pdata->rc.costbase - pdata->rc.optcost[1];
    pdata->rc.threshold[1] =
      TagIsSmall(lbt) ? pdata->rc.optcost[1] - GetSmall(lbt) :
      CLPFD_MAXINT-1;
  }
  if (pdata->rc.threshold[b]<0)	/* minimum is too large,
				   or maximum is too small */

    return FALSE;

  for (i=0; i<n; i++) {
    rowiter it;		/* iterator for one variable */
    rowiter_init(&it,i);
    nextj = rowiter_next(&it);
    for (j=0; j<n; j++) {
      if (j<nextj)
	rcost[n*i + j] = CLPFD_MAXINT;
      else {
	rcost[n*i + j] = reduced_cost(i,j);
	nextj = rowiter_empty(&it) ? n : rowiter_next(&it);
      }
    }
  }
  for (i=0; i<n; i++) {
    int mate = primal[0][i];
    for (j=0; j<n; j++)
      pdata->rc.dist[n*i + j] = rcost[n*j + mate];
  }
  all_pairs_shortest_paths(pdata,primal);
  max_regret = 0;
  min_regret = CLPFD_MAXINT;
  for (i=0; i<n; i++)
    for (j=0; j<n; j++) {
      long regret =
	safeadd(rcost[n*i + j],
		pdata->rc.dist[n*i + primal[1][j]]);
      rcost[n*i + j] = regret;
      if (max_regret<regret)
	max_regret = regret;
      if (regret>0 && min_regret>regret)
	min_regret = regret;
    }
  pdata->rc.min_regret[b] = min_regret;
  pdata->rc.max_regret[b] = max_regret;
  return TRUE;
}

/* Same algo. as single_circuit, but adapted for cost_based_pruning. */
static BOOL 
cost_cycle_feasible(struct all_dist_data *pdata,
		    long max_regret,
		    int *map,
		    int b)
{
  int id=0, top=0, min, m, vertid, i, vertex=0, child;
  long *lstack = (long *)pdata->dist_stack;
  long *regret = pdata->rc.rcost[b];
  int n = pdata->rc.ntargets;
  int *visited = pdata->rc.index;
  
  for (i=0; i<n; i++)
    visited[i] = 0;
start:
  visited[vertex] = vertid = min = ++id;
  for (i=0; i<n; i++) {
    if (regret[n*vertex + i]>max_regret)
      continue;
    child = map[i];
    m = visited[child];
    if (m==0) {
      lstack[top] = vertex;
      lstack[top+1] = i;
      lstack[top+2] = min;
      top += 3;
      vertex = child;
      goto start;
    }
  cont:
    if (min > m)
      min = m;
  }
  if (min==vertid)
    return (min==1 && id==n);
  m = min;
  top -= 3;
  vertex = lstack[top];
  i = lstack[top+1];
  min = lstack[top+2];
  vertid = visited[vertex];
				/* here, m is the minimal ID reachable from
				   some just visited child of vertex */
  goto cont;
}

#if 0 /* MAY COME IN HANDY */
/* Ripped off from visit() */
static int
cost_visit(struct all_dist_data *pdata,
	   int vertex,
	   long max_regret,
	   int *map,
	   int b)
{
  int newvertex=EOL;
  int i, m, min, scc;
  int *dist_stack = (int *)pdata->dist_stack;
  int n = pdata->rc.ntargets;
  int *visited = pdata->rc.index;
  long *regret = pdata->rc.rcost[b];
   
  visited[vertex] = ++pdata->scc_visited;
  min = pdata->scc_visited;
  dist_stack[pdata->scc_index++] = vertex;
  for (i=0; i<n; i++) {
    if (regret[n*vertex + i]>max_regret)
      continue;
    newvertex = map[i];
    if (newvertex != vertex) {
      m = (visited[newvertex] == 0) ? 
	cost_visit(pdata,newvertex,max_regret,map,b) :
	visited[newvertex];
      if (m < min)
	min = m;
    }
   }
   if (min == visited[vertex]) {
     scc = ++pdata->scc_component;
     do {
       /* Each descendant on the dist_stack is part of this vertex's SCC. */
       newvertex = dist_stack[--pdata->scc_index];
       COMPONENT(newvertex) = scc;
       visited[newvertex]=0xffffff; /* High value, so that
				       this vertex will be ignored
				       in the future search. */
     } while (newvertex != vertex);
   }
   return min;
}
#endif

static BOOL
cost_based_pruning MAGIC (HIDDEN_PROTO
			  struct all_dist_data *pdata,
			  int flags,
			  BOOL *change)
{
  BOOL cost_has_hole = !dvar_is_interval(pdata->rc.costvar);
  BOOL pruned = FALSE;
  long bound[2];
  int n = pdata->rc.ntargets;
  int nvars = pdata->nvars;
  int i, j, b;

  for (b=0; b<2; b++)
    bound[b] = pdata->rc.optcost[b];
  if (pdata->rc.max_regret[0] > pdata->rc.threshold[0] ||
      pdata->rc.max_regret[1] > pdata->rc.threshold[1] ||
      cost_has_hole) {
    long *threshold = pdata->rc.threshold;
    long **rcost = pdata->rc.rcost;
    long regret[2];
    
    /* PRUNE EACH VAR */
    for (i=0; i<n; i++) {
      VERTEX v = pdata->rc.index_var[i];

      if (LIVE(v)>1) {
	rowiter it;		/* iterator over variable v */
	FDCONS cons;		/* constructor for pruned values */
	
	rowiter_init(&it,i);
	fdcons_init(&cons);
	while (!rowiter_empty(&it)) {
	  j = rowiter_next(&it);
	  regret[0] = rcost[0][n*i + j];
	  regret[1] = rcost[1][n*i + j];
	  if (regret[0] > threshold[0] ||
	      regret[1] > threshold[1] ||
	      (cost_has_hole && regret[0]+regret[1]>0 &&
	       dvar_compare_interval_l(pdata->rc.costvar,
				       pdata->rc.optcost[0] + regret[0],
				       pdata->rc.optcost[1] - regret[1]) ==
	       FDI_DISJOINT))
	    fdcons_add(&cons,VAL(pdata->rc.index_var[j+n]));
	}
	if (fdcons_size(&cons)>0) {
	  if (dvar_prune_set(DVAR(v),fdcons_set(&cons)) == -1)
	    return FALSE;
	  LIVE(v) -= fdcons_size(&cons);
	  pruned = TRUE;
	}
      }
    }
    
    /* PRUNE EACH VAL */
    for (j=0; j<n; j++) {
      VERTEX v = pdata->rc.index_var[j+n];
      
      if (LIVE(v)>1) {
	coliter it;		/* iterator over value v */
	FDCONS cons;		/* constructor for pruned values */

	coliter_init(&it,j);
	fdcons_init(&cons);
	while (!coliter_empty(&it)) {
	  i = coliter_next(&it);
	  regret[0] = rcost[0][n*i + j];
	  regret[1] = rcost[1][n*i + j];
	  if (regret[0] > threshold[0] ||
	      regret[1] > threshold[1] ||
	      (cost_has_hole && regret[0]+regret[1]>0 &&
	       dvar_compare_interval_l(pdata->rc.costvar,
				       pdata->rc.optcost[0] + regret[0],
				       pdata->rc.optcost[1] - regret[1]) ==
	       FDI_DISJOINT))
	    fdcons_add(&cons,VAL(pdata->rc.index_var[i]));
	}
	if (fdcons_size(&cons)>0) {
	  if (dvar_prune_set(DVAR(v),fdcons_set(&cons)) == -1)
	    return FALSE;
	  LIVE(v) -= fdcons_size(&cons);
	  pruned = TRUE;
	}
      }
    }
  }
  

#if 0 /* MAY COME IN HANDY */
  /* CHECK PROPOSITION: all zero-regret solutions have cost=LB */
  {
    int val[10];
    int row = 0;
    int col = 0;

    val[0] = -1;
    while (row>=0) {
      col = val[row]+1;
      while (col<n && pdata->rc.rcost[0][n*row + col]>0)
	col++;
      if (col==n)
	row--;
      else {
	for (j=0; j<row && val[j]!=col; j++)
	  ;
	val[row] = col;
	if (j==row) {
	  if (row==n-1) {
	    long total = pdata->rc.costbase;
	    for (j=0; j<n; j++)
	      total -= pdata->rc.cost[n*j + val[j]]; /* rc.cost is negated */
	    if (total > pdata->rc.optcost[0]) {
	      printf("! PROPOSITION FALSE\n");
	      return FALSE;
	    }					       
	  } else
	    val[++row] = -1;
	}
      }
    }
  }
#endif

  if (flags&1)
    /* CHECK HAMILTONIAN FEASIBILITY */
    for (b=0; b<2; b++) {
      int nn = n*n;
      long *regarr = Malloc(nn,long);
      long *regret = pdata->rc.rcost[b];
      long threshold = pdata->rc.threshold[b];
      int nregrets = 0;
      int inf = -1, mid, sup;
      long optreg = 0;
      VERTEX v1, v2, *head, *tail;
      int *map = pdata->rc.heap;	/* val index -> var index */
      int *invmap = pdata->rc.vheap;	/* var index -> val index */
      int *index_var = pdata->rc.index_var;
      int *var_index = pdata->rc.var_index;
      
      head = pdata->dist_stack;
      tail = head + nvars;
      for (i=0; i<nvars; i++)
	head[i] = tail[i] = i;

      /* merge circuits for the "non rc.targets" */
      for (i=n; i<nvars; i++) {
	v1 = TVAR(i);
	v2 = dvar_min_l(DVAR(v1))-1;
	v1 = head[v1];
	v2 = tail[v2];
	tail[v1] = v2;
	head[v2] = v1;
      }
      for (i=0; i<n; i++) {
	map[i] = var_index[tail[index_var[i+n]-nvars]];
	invmap[map[i]] = i;
      }

#if 0 /* MAY COME IN HANDY */
      long regret0, regret1=0;
      int *comp_type = (int *)pdata->dist_stack;
      int *comp_minin = comp_type+n;
      int *comp_minout = comp_minin+n;
      int *visited = pdata->rc.index;
      int ncomp;
      /* INFER REGRETS FROM SCC ON SUBGRAPH BASED ON REGRET THRESHOLD */
      do {
	regret0 = regret1;

	pdata->scc_component = 0;
	pdata->scc_visited = 0;
	pdata->scc_index = 0;
	for (i=0; i<n; i++)
	  visited[i] = 0;
	for (i=0; i<n; i++)
	  if (visited[i] == 0)
	    cost_visit(pdata,i,regret0,map,b);
	ncomp = pdata->scc_component;
	if (ncomp==1)
	  break;
	for (i=0; i<ncomp; i++) {
	  comp_type[i] = 3;
	  comp_minin[i] = CLPFD_MAXINT;
	  comp_minout[i] = CLPFD_MAXINT;
	}
	for (i=0; i<nn; i++) {
	  long reg = regret[i];
	  int ifrom, ito, cfrom, cto;
      
	  if (reg<=threshold) {
	    ifrom = i/n;
	    ito = map[i%n];
	    cfrom = COMPONENT(ifrom)-1;
	    cto = COMPONENT(ito)-1;
	    if (cfrom!=cto) {
	      if (reg<=regret0) {
		comp_type[cfrom] &= -3; /* mark not SINK */
		comp_type[cto] &= -2; /* mark not SOURCE */
	      } else {
		if (comp_minout[cfrom]>reg)
		  comp_minout[cfrom] = reg;
		if (comp_minin[cto]>reg)
		  comp_minin[cto] = reg;
	      }
	    }
	  }
	}
	for (i=0; i<ncomp; i++) {
	  if ((comp_type[i]&1) && regret1<comp_minin[i]) /* SOURCE component */
	    regret1 = comp_minin[i];
	  if ((comp_type[i]&2) && regret1<comp_minout[i]) /* SINK component */
	    regret1 = comp_minout[i];
	}
      } while (regret0 < regret1);
#endif
      for (i=0, j=0; i<nn; i++)
	if (regret[i]<=threshold)
	  regarr[j++] = regret[i];
      nregrets = j;
      qsort_asc_long(regarr, nregrets);
      for (i=0, j=0; i<nregrets; i++)
	if (i==0 || regarr[i]>regarr[i-1])
	  regarr[j++] = regarr[i];
      nregrets = j;

      inf = 0;
      sup = nregrets;
      while (inf<sup) {
	mid = (inf+sup)>>1;
	if (!cost_cycle_feasible(pdata,regarr[mid],map,b))
	  inf = mid+1;
	else
	  sup = mid;
      }
      optreg = inf<nregrets ? regarr[inf] : CLPFD_MAXINT;
#if 0 /* MAY COME IN HANDY */
      /* Compute a min-cost MST with Kruskal's algorithm. */
      if (b==0 && n>2 && inf<nregrets) {
	long mstcost = pdata->rc.costbase;
	int *set = (int *)head;
	int *indexarr = (int *)regarr;
	int ncomp = n;
	int untaken = -1;
	int nedges;
	
	for (i=0; i<n; i++)
	  set[i] = i;		/* poor man's Union-Find */
	for (i=0, j=0; i<nn; i++) {
	  if (regret[i]<=optreg) {
	    long cost, altcost;
	    int u, v, altu, altv;

	    u = i/n;
	    v = i%n;
	    altu = map[v];
	    altv = invmap[u];
	    cost =  -pdata->rc.cost[i];
	    altcost =  -pdata->rc.cost[n*altu + altv];
	    if (cost<altcost || (cost==altcost && u<v))
	      indexarr[j++] = i;
	  }
	}
	nedges = j;
	qsort_asc_cost(indexarr, nedges);
	for (i=0; i<nedges && ncomp>1; i++) {
	  int u, v;
	  long cost;
	  
	  u = v = indexarr[i];
	  cost = -pdata->rc.cost[u]; /* rc.cost is negated */
	  u /= n;
	  v = map[v%n];
	  while (set[u]!=u) u = set[u];
	  while (set[v]!=v) v = set[v];
	  if (u==v) {
	    if (untaken == -1)
	      untaken = indexarr[i];
	  } else {
	    ncomp--;
	    mstcost += cost;
	    if (u<v) set[v] = u;
	    else     set[u] = v;
	  }
	}
	/* add cost of one more arc */
	if (untaken == -1 && i<nedges)
	  untaken = indexarr[i];
	if (untaken != -1)
	  mstcost -= pdata->rc.cost[untaken];/* rc.cost is negated */
      }
#endif
      Free(regarr);
      if (inf==nregrets)
	return FALSE;
      if (optreg!=0) {
	if (b==1) optreg = -optreg;
	bound[b] += optreg;
	pdata->rc.min_regret[b] = 0;
      }
    }

  /* PRUNE COST VAR */
  if (!dvar_contains_value_l(pdata->rc.costvar,bound[0])) {
    if (pdata->rc.min_regret[0]==CLPFD_MAXINT)
      return FALSE;
    bound[0] += pdata->rc.min_regret[0];
  }
  switch (dvar_fix_min_l(pdata->rc.costvar,bound[0])) {
  case -1:
    return FALSE;
  case 0:
    break;
  default:
    pruned = TRUE;
  }

  if (!dvar_contains_value_l(pdata->rc.costvar,bound[1])) {
    if (pdata->rc.min_regret[1]==CLPFD_MAXINT)
      return FALSE;
    bound[1] -= pdata->rc.min_regret[1];
  }
  switch (dvar_fix_max_l(pdata->rc.costvar,bound[1])) {
  case -1:
    return FALSE;
  case 0:
    break;
  default:
    pruned = TRUE;
  }

  pdata->pruned |= pruned;
  *change = pruned;
  return TRUE;
}


static BOOL cost_filter MAGIC (HIDDEN_PROTO
			       struct all_dist_data *pdata,
			       long costbase,
			       int flags)
{
  int nvars = pdata->nvars;
  int nvartargets = pdata->nvartargets;
  int nvaltargets = pdata->nvaltargets;
  VERTEX vertex;
  BOOL ok, change;
  int i, n;
  
  dvar_init(pdata->rc.costvar, RefAttr(nvars<<1), RefVar(nvars<<1));

  if (pdata->incremental)
    n = pdata->rc.ntargets;
  else {
    pdata->rc.costbase = costbase;
    for (i=0, n=0; i<nvartargets; i++) {
      vertex = TVAR(i);
      pdata->rc.var_index[vertex] = n;
      pdata->rc.index_var[n++] = vertex;
    }
    for (i=0; i<nvaltargets; i++) {
      VERTEX vertex = TVAL(i);
      pdata->rc.var_index[vertex] = n;
      pdata->rc.index_var[n++] = vertex;
    }
    n >>= 1;
    pdata->rc.ntargets = n;
  }
  pdata->rc.cost = Malloc(n*n,long); /* pos. or neg. submatrix */
  pdata->rc.dist = Malloc(n*n,long); /* distance matrix */
  pdata->rc.label[0] = Malloc(n,int);
  pdata->rc.label[1] = Malloc(n,int);
  pdata->rc.queue = Malloc(2*n,int);
  pdata->rc.index = Malloc(n,int);
  pdata->rc.heap = Malloc(n,int);
  pdata->rc.vheap = Malloc(n,int);
  ok = TRUE;
  do {
    change = FALSE;
    if (!cost_based_filter(pdata,0) ||
	!cost_based_filter(pdata,1) ||
	!cost_based_pruning(pdata,flags,&change))
      ok = FALSE;
  } while (ok && change);
  SP_free(pdata->rc.cost);
  SP_free(pdata->rc.dist);
  SP_free(pdata->rc.label[0]);
  SP_free(pdata->rc.label[1]);
  SP_free(pdata->rc.queue);
  SP_free(pdata->rc.index);
  SP_free(pdata->rc.heap);
  SP_free(pdata->rc.vheap);
  return ok;
}

void SPCDECL 
prolog_fd_assignment_helper MAGIC (HIDDEN_PROTO
				   SP_term_ref State0,
				   SP_term_ref State,
				   SP_term_ref Actions)
{
  WAMENV;
  int nvars, n, ix, ival;
  TAGGED tstamp, thandle, tval;
  long lreg, ureg;
  int ent = 1;
  struct all_dist_data *pdata;
  Dvar dv;

  (void)State0;                 /* [PM] 3.9b5 avoid -Wunused */
  w->numstack_end = NULL;
  dvar_export_start(3);

/*    X(0) = RefTerm(State0); */
  DerefNonvar(X(0));
  RefTerm(State) = X(0);	/* f(Index,Var,StateMutable) */
  DerefArg(tval,X(0),2);
  if (IsVar(tval)) {
    ent = 0; goto ret;
  }
  ix = GetSmall(CTagToArg(X(0),1));
  ival = GetSmall(tval);
  thandle = RefMutable(CTagToArg(X(0),3));
  DerefArg(tstamp,thandle,ASSIGNMENT_ARITY);
  DerefArg(thandle,thandle,ASSIGNMENT_ARITY-1);
  if (thandle==TaggedZero)
    goto ret;
  pdata = Pdata(struct all_dist_data,thandle);
  if (GetSmall(tstamp)!=pdata->stamp)
    goto ret;
  n = pdata->rc.ntargets;
  nvars = pdata->nvars;
  if (VISITED(ix)==EOL)
    goto ret;			/* nothing to do */
  ix = pdata->rc.var_index[ix];
  ival = pdata->rc.var_index[ival+nvars-1]-n;
  lreg = pdata->rc.rcost[0][n*ix + ival];
  ureg = pdata->rc.rcost[1][n*ix + ival];
  if (lreg==CLPFD_MAXINT || ureg==CLPFD_MAXINT) {
    ent = -1; goto ret;
  }
  dv = pdata->rc.costvar;
  dvar_init(dv, RefAttr(nvars<<1), RefVar(nvars<<1));
  if (dvar_fix_interval_l(dv, pdata->rc.optcost[0]+lreg,
			      pdata->rc.optcost[1]-ureg) == -1) {
    ent = -1; goto ret;
  }
  dvar_pruning_done(dv);
  dvar_export(dv, 3);
 ret:
  dvar_export_done(Actions, ent, 3);
}
