/* Copyright(C) 2000, Swedish Institute of Computer Science */

#include "fd.h"
#include "dvars.h"

#if MULTI_SP_AWARE
#if 0                           /* not redefined. It uses FD_SETUP_SPENV to access the SPEnv */
#define case_destructor(A1) case_destructor(HIDDEN_ARG, A1)
#endif
#define eval_leaf(A1,A2) eval_leaf(HIDDEN_ARG, A1,A2)
#define eval_dag(A1,A2,A3) eval_dag(HIDDEN_ARG, A1,A2,A3)
#endif /* MULTI_SP_AWARE */


#define CASE_DOM 0
#define CASE_MINMAX 1
#define CASE_MIN 2
#define CASE_MAX 3
#define CASE_VAL 4
#define CASE_NONE 5

struct case_data {
  void (SPCDECL *destructor)(void*);
#if MULTI_SP_AWARE
  SPEnv *spenv;
#endif /* MULTI_SP_AWARE */

  SP_globref refbase;
  long stamp;			/* increases up to backtracking */
  int use_leaf;
  int nvars;
  int nnodes;
  unsigned long false_date;
  unsigned long true_date;
  TAGGED relevant_vars;		/* volatile, numstack data */
  Dvar dvar;
  TAGGED *prune_set;		/* for DOM, volatile, GC-unsafe */
  TAGGED *prune_min;		/* for MINMAX|MIN|VAL, volatile, atomic */
  TAGGED *prune_max;		/* for MINMAX|MAX|VAL, volatile, atomic */
  long   *prune_mask;		/* any=1, min=2, max=4, at least two=8, volatile */
  unsigned long *control;
  struct {
    unsigned long *date;	/* TRUE(FALSE) if = true_date(false_date), undefined otherwise */
    long *id;			/* unique by definition */
    long *var_index;
    long *child_index;
    long *child_end;
  } dag_node;
  struct {
    TAGGED *min, *max;
    long *node_index;
  } child;
};

#define DATE(N) (pdata->dag_node.date[N])
#define ID(N) (pdata->dag_node.id[N])
#define VAR(N) (pdata->dag_node.var_index[N])
#define VARSET(N) RefGlob(pdata->refbase + 2*nvars + (N))
#define CHILD(N) (pdata->dag_node.child_index[N])
#define CHILD_END(N) (pdata->dag_node.child_end[N])
#define CMIN(C) (pdata->child.min[C])
#define CMAX(C) (pdata->child.max[C])
#define NODE(C) (pdata->child.node_index[C])

static void SPCDECL case_destructor(void *pdata_v)
{
  struct case_data *pdata = (struct case_data *)pdata_v;
  FD_SETUP_SPENV(pdata->spenv)

  SP_free_globrefs(pdata->refbase,2*pdata->nvars+pdata->nnodes);
  SP_free(pdata);
}


static int MAGIC eval_leaf MAGIC (HIDDEN_PROTO
				  long id,
				  struct case_data *pdata)
{
  int nvars = pdata->nvars;
  int virtual = nvars-1;
  int ctrl;
  TAGGED tvirtual = MakeSmall(virtual);
  TAGGED tid = MakeSmall(id);
  TAGGED dom, set, min, max;
    
  DerefAttribute(dom,RefGlob(pdata->refbase+(virtual<<1))); /* dom/4 term */
  set = DomainSet(dom);
  min = DomainMin(dom);
  max = DomainMax(dom);
    
  if (InInterval(tid,min,max) && fd_member(tid,set)) { /* feasible */
    if (fd_member(tvirtual,pdata->relevant_vars)) {
      pdata->prune_mask[virtual] |= 0x1; /* support for some domain element - see CASE_VAL */
      if (min==tid)
	pdata->prune_mask[virtual] |= 0x2;	/* support for min element */
      if (max==tid)
	pdata->prune_mask[virtual] |= 0x4;	/* support for max element */
      switch ((ctrl=pdata->control[virtual])) {
      case CASE_DOM:
	{
	  TAGGED set1;
	  pdata->prune_set[virtual] = set1 =
	    fd_delete(pdata->prune_set[virtual],tid);
	  if (set1==EmptySet)
	    pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	}
	break;
      case CASE_MIN:
	if (pdata->prune_mask[virtual] & 0x2)
	  pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	else {
	  if (FDgt(pdata->prune_min[virtual],tid))
	    pdata->prune_min[virtual] = tid;
	}
	break;
      case CASE_MAX:
	if (pdata->prune_mask[virtual] & 0x4)
	  pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	else {
	  if (FDlt(pdata->prune_max[virtual],tid))
	    pdata->prune_max[virtual] = tid;
	}
	break;
      case CASE_MINMAX:
      case CASE_VAL:
	if ((pdata->prune_mask[virtual] & 0x6) == 0x6)
	  pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	else {		/* detect when we have support for at least two values */
	  if (FDgt(pdata->prune_min[virtual],tid))
	    pdata->prune_min[virtual] = tid;
	  if (FDlt(pdata->prune_max[virtual],tid))
	    pdata->prune_max[virtual] = tid;
	  if (ctrl==CASE_VAL &&
	      pdata->prune_min[virtual] != pdata->prune_max[virtual])
	    pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	}
	break;
      }
    }
    return TRUE;
  } else { /* infeasible */
    return FALSE;
  }
}

static int MAGIC eval_dag MAGIC (HIDDEN_PROTO Argdecl,
				 int node,
				 struct case_data *pdata)
{
  int state = FALSE;
  int leaf_state = TRUE;
  int use_leaf = pdata->use_leaf;
  
  if (DATE(node) == pdata->true_date)
    return TRUE;
  if (DATE(node) == pdata->false_date)
    return FALSE;
  {
    int virtual = VAR(node);
    TAGGED tvirtual = MakeSmall(virtual);
    int child_index = CHILD(node);
    int child_end = CHILD_END(node);
    int nvars = pdata->nvars;
    int inf, mid, sup;		/* for dichotomic search */
    int child;
    TAGGED dom, set, min, max;
    
    DerefAttribute(dom,RefGlob(pdata->refbase+(virtual<<1))); /* dom/4 term */
    set = DomainSet(dom);
    min = DomainMin(dom);
    max = DomainMax(dom);
    
    /* dichotomic search for first child */
    inf = child_index;
    sup = child_end;
    while (inf<sup) {
      mid = (inf+sup)>>1;
      if (FDgt(min,CMAX(mid))) /* current variable #> child */
	inf = mid+1;
      else
	sup = mid;
    }
    child = child_index = inf;
    
    /* scan the compatible children */
    while (child_index < child_end &&
	   leaf_state &&
	   !FDgt(CMIN(child),max) &&
	   (!state ||
	    fd_compare(VARSET(node),pdata->relevant_vars)!=FDI_DISJOINT)) {
      int lstate = TRUE;

      if (CMIN(child)==CMAX(child)) { /* singleton interval special case */
	if (fd_member(CMIN(child),set))
	  goto intersect;
	else
	  goto disjoint;
      }
      switch (fd_compare_interval(set,CMIN(child),CMAX(child))) {
      case FDI_DISJOINT:
      disjoint:
	lstate = FALSE;
	break;
      case FDI_SUBSET:
      case FDI_EQUAL:
	if (NODE(child) > -1) {
	  /* TODO: assert RefGlob(pdata->refbase+(virtual<<1)+1) in CMIN(child) .. CMAX(child) */
	  lstate = eval_dag(w,NODE(child),pdata);
	  /* TODO: retract RefGlob(pdata->refbase+(virtual<<1)+1) in CMIN(child) .. CMAX(child) */
	} else if (use_leaf-- > 0)
	  lstate = leaf_state = eval_leaf(ID(node),pdata);
	if (lstate && fd_member(tvirtual,pdata->relevant_vars))
	  pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	break;
      default:
      intersect:
	if (NODE(child) > -1) {
	  /* TODO: assert RefGlob(pdata->refbase+(virtual<<1)+1) in CMIN(child) .. CMAX(child) */
	  lstate = eval_dag(w,NODE(child),pdata);
	  /* TODO: retract RefGlob(pdata->refbase+(virtual<<1)+1) in CMIN(child) .. CMAX(child) */
	} else if (use_leaf-- > 0)
	  lstate = leaf_state = eval_leaf(ID(node),pdata);
	if (lstate && fd_member(tvirtual,pdata->relevant_vars)) {
	  pdata->prune_mask[virtual] |= 0x1; /* support for some domain element - see CASE_VAL */
	  if (!FDgt(CMIN(child),min)) /* min >= CMIN(child) */
	    pdata->prune_mask[virtual] |= 0x2;	/* support for min element */
	  if (!FDgt(max,CMAX(child))) /* max <= CMAX(child) */
	    pdata->prune_mask[virtual] |= 0x4;	/* support for max element */
	  switch (pdata->control[virtual]) {
	  case CASE_DOM:
	    {
	      TAGGED set1;
	      pdata->prune_set[virtual] = set1 =
		fd_subtract_interval(pdata->prune_set[virtual],CMIN(child),CMAX(child));
	      if (set1==EmptySet)
		pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	    }
	    break;
	  case CASE_MIN:
	    if (pdata->prune_mask[virtual] & 0x2)
	      pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	    else {
	      if (FDgt(pdata->prune_min[virtual],CMIN(child)))
		pdata->prune_min[virtual] = CMIN(child);
	    }
	    break;
	  case CASE_MAX:
	    if (pdata->prune_mask[virtual] & 0x4)
	      pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	    else {
	      if (FDlt(pdata->prune_max[virtual],CMAX(child)))
		pdata->prune_max[virtual] = CMAX(child);
	    }
	    break;
	  case CASE_MINMAX:
	    if ((pdata->prune_mask[virtual] & 0x6) == 0x6)
	      pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	    else {
	      if (FDgt(pdata->prune_min[virtual],CMIN(child)))
		pdata->prune_min[virtual] = CMIN(child);
	      if (FDlt(pdata->prune_max[virtual],CMAX(child)))
		pdata->prune_max[virtual] = CMAX(child);
	    }
	    break;
	  case CASE_VAL:
	    if ((pdata->prune_mask[virtual] & 0x6) == 0x6)
	      pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	    else {		/* detect when we have support for at least two values */
	      /* TODO: fd_compare case analysis */
	      TAGGED set1 = fd_and_interval(set,CMIN(child),CMAX(child));
	      TAGGED min = fd_min(set1);
	      TAGGED max = fd_max(set1);
	      
	      if (FDgt(pdata->prune_min[virtual],min))
		pdata->prune_min[virtual] = min;
	      if (FDlt(pdata->prune_max[virtual],max))
		pdata->prune_max[virtual] = max;
	      if (pdata->prune_min[virtual] != pdata->prune_max[virtual])
		pdata->relevant_vars = fd_delete(pdata->relevant_vars,tvirtual);
	    }
	    break;
	  }
	}
      }
      /* update state of current node */
      state |= lstate;
      child_index++;
      child++;
    }
  }
  /* record state */
  DATE(node) = (state ? pdata->true_date : pdata->false_date);
  return state;
}

/*
   '$fd_case'(+State0, -State, -Actions) :-
   State0 = State = state(f(NVars,NNodes,_NChildren,TVars,Dag,Prune,UseLeaf),_Handle,Stamp)
 */
void SPCDECL
prolog_fd_case MAGIC (HIDDEN_PROTO
		      SP_term_ref State0,
		      SP_term_ref State,
		      SP_term_ref Actions)
{
  WAMENV;
  struct case_data *pdata;
  int i, nvars, nnodes, nchildren, nnonground, ent = -1;
  TAGGED item, tmp, state, handle;
  long state_stamp;
  BOOL committed;		/* not used */
  char *ptr;
  FDITER it;
  
  w->numstack_end = NULL;

/*    X(0) = RefTerm(State0); */
  (void)State0;
  X(2) = atom_nil;		/* actions list */
  RefTerm(State) = unify_output_state(w,&handle,&state_stamp,&committed);

  if (!IsVar(handle)) {		/* got [Flag | '$free'(Ptr)] */
    pdata = Pdata(struct case_data,handle);
    nvars = pdata->nvars;
    nnodes = pdata->nnodes;
    pdata->stamp = state_stamp-1; /* DON'T trust initially */
  } else {			/* build persistent state */
    int child;
    TAGGED tvars, dag, prune;
    
    DerefArg(state,X(0),1);
    DerefArg(tmp,state,1); 
    nvars = GetSmall(tmp);
    DerefArg(tmp,state,2); 
    nnodes = GetSmall(tmp);
    DerefArg(tmp,state,3); 
    nchildren = GetSmall(tmp);
    pdata = Palloc(struct case_data,
		   nvars*sizeof(struct dvar)+
		   nvars*sizeof(long)+
		   nnodes*5*sizeof(long)+
		   nchildren*3*sizeof(long),
		   handle);
    pdata->destructor = case_destructor;
    FD_STORE_SPENV(pdata->spenv);
    pdata->refbase = SP_alloc_globrefs(2*nvars+nnodes);
    pdata->nvars = nvars;
    pdata->nnodes = nnodes;
    pdata->false_date = 0L;
    pdata->true_date = 1L;
    ptr = (char *)(pdata+1);
    pdata->dvar = (Dvar)ptr; ptr += nvars*sizeof(struct dvar);
    pdata->control = (unsigned long *)ptr; ptr += nvars*sizeof(long);
    pdata->dag_node.date = (unsigned long *)ptr; ptr += nnodes*sizeof(long);
    pdata->dag_node.id = (long *)ptr; ptr += nnodes*sizeof(long);
    pdata->dag_node.var_index = (long *)ptr; ptr += nnodes*sizeof(long);
    pdata->dag_node.child_index = (long *)ptr; ptr += nnodes*sizeof(long);
    pdata->dag_node.child_end = (long *)ptr; ptr += nnodes*sizeof(long);
    pdata->child.min = (TAGGED *)ptr; ptr += nchildren*sizeof(long);
    pdata->child.max = (TAGGED *)ptr; ptr += nchildren*sizeof(long);
    pdata->child.node_index = (long *)ptr; ptr += nchildren*sizeof(long);
    
    DerefArg(tvars,state,4); 
    DerefArg(dag,state,5); 
    DerefArg(prune,state,6); 
    for (i=0; i < nvars; i++) {
      SP_globref ref = pdata->refbase+(i<<1);

      DerefCar(item,tvars);
      DerefCdr(tvars,tvars);
      DerefArg(tmp,item,1);	/* var's domain variable */
      RefGlob(ref+1) = tmp;
      DerefArg(tmp,item,2);	/* var's attribute */
      RefGlob(ref) = tmp;
    }
    DerefArg(tmp,state,7);
    pdata->use_leaf = GetSmall(tmp);
    /* build the DAG */
    child = 0;
    for (i=0; i < nnodes; i++) {
      int node = i;
      TAGGED children, dom, cmin, cmax;
      int ix;

      DerefCar(item,dag);	/* get dagnode(ID,VAR,VARSET,Children) */
      DerefCdr(dag,dag);
      DerefArg(tmp,item,1);	/* node ID */
      ID(node) = GetSmall(tmp);
      DerefArg(tmp,item,2);	/* var index */
      DATE(node) = -1L;
      VAR(node) = ix = GetSmall(tmp);
      DerefAttribute(dom,RefGlob(pdata->refbase+(ix<<1))); /* dom/4 term */
      DerefArg(tmp,item,3);	/* var set */
      VARSET(node) = tmp;
      CHILD(node) = child;
      DerefArg(children,item,4); /* children */
      while (TagIsLST(children)) {
	DerefCar(item,children);
	DerefCdr(children,children);
	DerefArg(tmp,item,2);
	DerefArg(item,item,1);
	DerefArg(cmin,item,1);
	DerefArg(cmax,item,2);
	CMIN(child) = cmin;
	CMAX(child) = cmax;
	NODE(child) = (TagIsSmall(tmp) ? GetSmall(tmp) : -1);
	child++;
      }
      CHILD_END(node) = child;
    }
    /* build control info */
    for (i=0; i<nvars; i++) {
      DerefCar(item,prune);
      DerefCdr(prune,prune);
      item = SetArity(item,1);
      if (item==fd.functor_dom)
	pdata->control[i] = CASE_DOM;
      else if (item==fd.functor_min)
	pdata->control[i] = CASE_MIN;
      else if (item==fd.functor_max)
	pdata->control[i] = CASE_MAX;
      else if (item==fd.functor_minmax)
	pdata->control[i] = CASE_MINMAX;
      else if (item==fd.functor_val)
	pdata->control[i] = CASE_VAL;
      else
	pdata->control[i] = CASE_NONE;
    }
  }
  
  /* RESUME */
  ptr = SP_malloc(4*nvars*sizeof(TAGGED));
  pdata->prune_set = (TAGGED *)ptr; ptr += nvars*sizeof(TAGGED);
  pdata->prune_min = (TAGGED *)ptr; ptr += nvars*sizeof(TAGGED);
  pdata->prune_max = (TAGGED *)ptr; ptr += nvars*sizeof(TAGGED);
  pdata->prune_mask = (long *)ptr;  ptr += nvars*sizeof(long);
  if (state_stamp != pdata->stamp) { /* all nodes must be recomputed */
    pdata->true_date += 2;
    pdata->false_date += 2;
  } else {			/* false nodes remain false */
    pdata->true_date += 2;
  }
  pdata->stamp = state_stamp+1;

  /* init each time */
  pdata->relevant_vars = EmptySet;
  nnonground = 0;
  for (i=0; i < nvars; i++) {
    SP_globref ref = pdata->refbase+(i<<1);
    Dvar dv = pdata->dvar+i;
    
    dvar_init(dv, ref, ref+1);
    if (pdata->control[i]==CASE_DOM)
      pdata->prune_set[i] = dvar_set(dv);
    else {
      pdata->prune_min[i] = Sup; /* smallest so far */
      pdata->prune_max[i] = Inf; /* greatest so far */
    }
    pdata->prune_mask[i] = 0;
    if (!dvar_is_integer(dv)) {
      if (pdata->control[i]!=CASE_NONE)
	pdata->relevant_vars =
	  fd_insert_into(MakeSmall(i),pdata->relevant_vars);
      nnonground++;
    }
  }
  /* evaluate the DAG */
  if (!eval_dag(w,0,pdata))
    goto ret;

  /* compute pruning acc. to generalization */

  fditer_init(&it, pdata->relevant_vars);
  while (!fditer_empty(&it)) {
    Dvar dv;
    
    tmp = fditer_next(&it);
    i = GetSmall(tmp);
    dv = pdata->dvar+i;
    switch (pdata->control[i]) {
    case CASE_DOM:
      dvar_prune_set(dv, pdata->prune_set[i]);
      break;
    case CASE_MIN:
      dvar_fix_min_t(dv, pdata->prune_min[i]);
      break;
    case CASE_MAX:
      dvar_fix_max_t(dv, pdata->prune_max[i]);
      pdata->prune_set[i] = fd_interval(Inf,pdata->prune_max[i]);
      break;
    case CASE_MINMAX:
    case CASE_VAL:
      dvar_fix_interval_t(dv, pdata->prune_min[i], pdata->prune_max[i]);
      break;
    }
    dvar_pruning_done(dv);
  }

  /* do actual pruning */
  fditer_init(&it, pdata->relevant_vars);
  while (!fditer_empty(&it)) {
    Dvar dv;
    
    tmp = fditer_next(&it);
    i = GetSmall(tmp);
    dv = pdata->dvar+i;
    dvar_export(dv, 3);
    if (dvar_is_integer(dv))
      nnonground--;
  }
  ent = !nnonground;
 ret:
  SP_free(pdata->prune_set);
  Pfree(ent==1,handle,pdata);
  dvar_export_done(Actions, ent, 3);
}


