/*
 * SICSTUS CLPFD DEMONSTRATION PROGRAM
 * Purpose   : Social Golfer Problem
 * Author    : Mats Carlsson
 *
 * We have 32 golfers, individual play.
 * We will golf for W weeks.
 * Set up the foursomes so that each person only golfs with the same
 * person once.
 */

% Best luck so far:
% | ?- golf(8,4,7,[ff],bycol).

:- module(golf, [golf/5]).

:- use_module(library(lists)).
:- use_module(library(clpfd)).

golf(G, S, W, Opt, VarOrder) :-
	golfer(G, S, W, Schedule, Byrow, Bycol),
	var_order(VarOrder, Byrow, Bycol, All),
	label_sets(All, Opt),
	display_rounds(Schedule, 0).

var_order(byrow, All, _, All).
var_order(bycol, _, All, All).

label_sets([], _).
label_sets([Set|Sets], Opt) :-
	labeling(Opt, Set),
	label_sets(Sets, Opt).

display_rounds([], _).
display_rounds([Round|Rounds], V) :-
	W is V+1,
	format('Week ~d:\n', [W]),
	display_round(Round),
	display_rounds(Rounds, W).

display_round([]).
display_round([Four|Round]) :-
	format('                    ~d ~d ~d ~d\n', Four),
	display_round(Round).

/* golfer(Number of Groups, GroupSize, NumberOfWeeks, ...) */
golfer(G, S, W, Schedule, PlayersByRow, PlayersByCol) :-
	schedule(0, G, S, W, Schedule, PlayersByRow, PlayersByCol),
	Schedule = [FirstS|RestS],
	concat(FirstS, Players, []),
	labeling([enum], Players), !,
	seed_rest(RestS, S),
	ordered_players_by_week(PlayersByRow),
	players_meet_disjoint(Schedule),
	first_s_alldiff(0, S, RestS).

schedule(W, _, _, W, [], [], []) :- !.
schedule(I, G, S, W, [Week|Schedule], [ByRow|ByRows], [ByCol|ByCols]) :-
	week(0, G, S, Week),
	concat(Week, ByRow, []),
	all_different(ByRow,[on(dom),consistency(global)]),
	transpose(Week, WeekT, WeekT),
	concat(WeekT, ByCol, []),
	J is I+1,
	schedule(J, G, S, W, Schedule, ByRows, ByCols).

week(G, G, _, []) :- !.
week(I, G, S, [Group|Week]) :-
	length(Group, S),
	GS is G*S-1,
	domain(Group, 0, GS),
	J is I+1,
	week(J, G, S, Week).

players_meet_disjoint(Schedule) :-
	concat(Schedule, Groups, []),
	groups_meets(Groups, Meets, []),
	disjoint2(Meets).

groups_meets([]) --> [].
groups_meets([Group|Groups]) -->
	group_meets(Group),
	groups_meets(Groups).

group_meets([]) --> [].
group_meets([P|Ps]) -->
	group_meets(Ps, P),
	group_meets(Ps).

group_meets([], _) --> [].
group_meets([Q|Qs], P) --> [r(P,1,Q,1)],
	group_meets(Qs, P).

concat([]) --> [].
concat([X|Xs]) --> concat1(X), concat(Xs).

concat1([]) --> [].
concat1([X|Xs]) --> [X], concat1(Xs).

seed_rest([], _).
seed_rest([Week|Rest], S) :-
	seed_week(0, S, Week),
	ascending_quotients(Week, S), % reduces #backtracks by 50%
	seed_rest(Rest, S).

seed_week(S, S, Week) :- !,
	S1 is S-1,
	seed_week(Week, S1).
seed_week(I, S, [[I|_]|Week]) :-
	J is I+1,
	seed_week(J, S, Week).

seed_week([], _).
seed_week([[J|_]|Week], I) :-
	I #< J,
	seed_week(Week, J).

ascending_quotients([], _).
ascending_quotients([Group|Groups], S) :-
	ascending_quotient(Group, S),
	ascending_quotients(Groups, S).

ascending_quotient([P|Ps], S) :-
	P/S #= Q,
	ascending_quotient(Ps, Q, S).

ascending_quotient([], _, _).
ascending_quotient([P|Ps], Q0, S) :-
	P/S #= Q,
	Q0 #< Q,
	ascending_quotient(Ps, Q, S).

transpose([], Zs, Xs) :-
	transpose_1(Zs, Xs).
transpose([Y|Ys], Zs, Xs) :-
	transpose_1(Xs, Y, Xs1),
	transpose(Ys, Zs, Xs1).

transpose_1([], []).
transpose_1([_|Zs], [[]|Xs]) :-
	transpose_1(Zs, Xs).

transpose_1([], [], []).
transpose_1([[H|T]|Xs], [H|Hs], [T|Ts]) :-
	transpose_1(Xs, Hs, Ts).

ordered_players_by_week([W|Ws]) :-
	ordered_players_by_week(Ws, W).

ordered_players_by_week([], _).
ordered_players_by_week([W|Ws], V) :-
	W = [_,Y|_],
	V = [_,X|_],
	X #< Y,
	ordered_players_by_week(Ws, W).

first_s_alldiff(S, S, _Schedule) :- !.
first_s_alldiff(I, S, Schedule) :-
	concat_ith(Schedule, I, Conc, []),
	all_different(Conc,[on(dom),consistency(global)]),
	J is I+1,
	first_s_alldiff(J, S, Schedule).

concat_ith([], _) --> [].
concat_ith([Week|S], I) -->
	{nth0(I, Week, [_|Ps])},
	concat1(Ps),
	concat_ith(S, I).

end_of_file.

Occurrences in the four columns:
W weeks, G groups, P players, foursomes.

            Column 1        Column 2            Column 3        Column 4
            ********        ********            ********        ********

 0..3          W-1              0               Rev(Column 2)   Rev(Column 1)
 4..7    W-4..(P-8)/3     0..min(4,(P-8)/2)     
 8..11   W-8..(P-12)/3    0..min(8,(P-12)/2)    
12..15  W-12..(P-16)/3    0..min(12,(P-16)/2)   
16..19  W-16..(P-20)/3    0..min(16,(P-20)/2)   
20..23        0           0..min(20,(P-24)/2)   
24..27        0                 0               
28..31        0                 0               

